import json
from google.protobuf.json_format import MessageToDict
import torch as th
import numpy as np
import onnx
import onnxruntime as ort
from typing import Tuple
from pathlib import Path

from naeural_core.serving.base.backends.model_backend_wrapper import ModelBackendWrapper

def onnx_to_np_dtype(onnx_type : str) -> np.dtype:
  onnx_to_np_dtype_dict = {
    "tensor(bool)": np.bool_,
    "tensor(int)": np.int32,
    'tensor(int32)': np.int32,
    'tensor(int8)': np.int8,
    'tensor(uint8)': np.uint8,
    'tensor(int16)': np.int16,
    'tensor(uint16)': np.uint16,
    'tensor(uint64)': np.uint64,
    "tensor(int64)": np.int64,
    'tensor(float16)': np.float16,
    "tensor(float)": np.float32,
    'tensor(double)': np.float64,
  }
  return onnx_to_np_dtype_dict[onnx_type]


def get_torch_dtype(np_dtype : np.dtype) -> th.dtype:
  """
  Get a torch dtype from a numpy dtype.
  Parameters:
    - np_dtype a numpy dtype
  Returns:
    - the corresponding torch dtype.
  """
  numpy_to_torch_dtype_dict = {
      bool          : th.bool,
      np.bool_      : th.bool,
      np.uint8      : th.uint8,
      np.int8       : th.int8,
      np.int16      : th.int16,
      np.int32      : th.int32,
      np.int64      : th.int64,
      np.float16    : th.float16,
      np.float32    : th.float32,
      np.float64    : th.float64,
      np.complex64  : th.complex64,
      np.complex128 : th.complex128
  }
  if np_dtype not in numpy_to_torch_dtype_dict.keys():
    raise ValueError("No known corresponding torch dtype for {}".format(np_dtype))
  return numpy_to_torch_dtype_dict[np_dtype]

def get_static_shape(shape, batch_size):
  # Replace all dynamic dimensions with batch size
  return [batch_size if isinstance(x, str) else x for x in shape]

# Wrapper around ONNX for inference.
class ONNXModel(ModelBackendWrapper):
  ONNX_METADATA_KEY = 'onnx_metadata'

  def __init__(self):
    self._session = None
    self._max_batch_size = None
    self._batched_input_idx = None
    self._batch_runtime_info = None
    self._metadata = None
    return


  def _get_batch_dependent_arg_idx(self) -> int:
    for i, input in enumerate(self._session.get_inputs()):
      shape = input.shape
      if isinstance(shape[0], str):
        return i
    # No batching?
    raise Exception("No batching dimension found")


  def _get_batch_runtime_state(self, batch_size : int):
    # Return a tuple of (list of (input_name, input_dtype, input_shape),
    # (list of (output shape, output name, numpy dtype, th dtype).
    inputs_info = []
    for input in self._session.get_inputs():
      input_name = input.name
      input_dtype = onnx_to_np_dtype(input.type)
      input_shape = get_static_shape(input.shape, batch_size)
      inputs_info.append((input_name, input_dtype, input_shape))
    #endfor all inputs

    # Produce the list of output tensor information for this batch size.
    outputs = []
    for output in self._session.get_outputs():
      output_shape = get_static_shape(output.shape, batch_size)
      np_dtype = onnx_to_np_dtype(output.type)
      th_dtype = get_torch_dtype(np_dtype)
      outputs.append((output_shape, output.name, np_dtype, th_dtype))
    #endfor outputs
    return inputs_info, outputs


  def load_model(
    self,
    model_path : str,
    max_batch_size : int,
    device : th.device,
    half : bool
  ) -> None:
    """
    Initialize from a path on the disk.

    Parameters
    ----------
      model_path - path to onnx model on disk
      max_batch_size - maximum batch size for the inference
      device - torch device to use.
      half - if True converts model to fp16

    Returns
    -------
    None
    """
    self._max_batch_size = max_batch_size
    self._device_type = device.type
    self._device_id = 0 if device.index is None else device.index
    self._device = device

    # Read the metadata from the model.
    model_onnx = onnx.load(model_path)
    onnx_metadata_dict = MessageToDict(model_onnx, preserving_proto_field_name = True)['metadata_props']
    for meta in onnx_metadata_dict:
      if meta['key'] == ONNXModel.ONNX_METADATA_KEY:
        self._metadata = json.loads(meta['value'])

    model_precision = self._metadata.get('precision')
    if model_precision is not None:
      if (model_precision.lower() == "fp16") != half:
        if half and model_precision.lower() == "fp32":
          #from onnxconverter_common import float16
          from onnxruntime.transformers import float16
          model_fp16 = float16.convert_float_to_float16(model_onnx, disable_shape_infer=True)
          fp16path = str(Path(model_path).with_suffix('.tofp16.onnx'))
          onnx.save(model_fp16, fp16path)
          model_path = fp16path
        else:
          raise RuntimeError('Incompatible precision in ONNX model')
    #endif check onnx precision

    del model_onnx
    model_onnx = None

    providers = ['CPUExecutionProvider']
    if device.type.startswith('cuda'):
      providers.append('CUDAExecutionProvider')

    self._session = ort.InferenceSession(
      model_path,
      providers=providers
    )
    self._input_dtypes = []
    for input in self._session.get_inputs():
      np_dtype = onnx_to_np_dtype(input.type)
      self._input_dtypes.append(np_dtype)

    self._batched_input_idx = self._get_batch_dependent_arg_idx()

    # Pre-construct shapes and output information for all possible batch sizes.
    # This avoids doing some work at inference time.
    self._batch_runtime_info = []
    for batch_size in range(self._max_batch_size + 1):
      self._batch_runtime_info.append(self._get_batch_runtime_state(batch_size))

    return


  def forward(self, *args : th.tensor) -> Tuple[th.tensor, ...]:
    """
    Run inference. All model inputs and outputs are torch tensors.
    Arguments are the input tensors in the order specified when the model
    was exported from torch, first inputs and then outputs (see
    create_from_torch).
    Users must pass all input tensors as unnamed arguments.
    Returns a tuple of output tensors, in the order of outputs of the
    original torch model.
    """
    batch_size = args[self._batched_input_idx].shape[0]
    binding = self._session.io_binding()
    input_info, output_info = self._batch_runtime_info[batch_size]

    for idx, (input_name, input_dtype, input_shape) in enumerate(input_info):
      binding.bind_input(
        name=input_name,
        device_type=self._device_type,
        device_id=self._device_id,
        element_type=input_dtype,
        shape=input_shape,
        buffer_ptr=args[idx].contiguous().data_ptr(),
      )
    #endfor all inputs
    outputs = []
    # Allocate and bind output tensors.
    for shape, name, np_dtype, th_dtype in output_info:
      th_output = th.empty(
        size=tuple(shape),
        dtype=th_dtype,
        device=self._device,
        requires_grad=False
      )
      outputs.append(th_output)
      binding.bind_output(
          name=name,
          device_type=self._device_type,
          device_id=self._device_id,
          element_type=np_dtype,
          shape=tuple(shape),
          buffer_ptr=th_output.data_ptr(),
      )
    #endfor all outputs

    if batch_size > 0:
      self._session.run_with_iobinding(binding)
    if len(outputs) == 1:
      return outputs[0]
    return tuple(outputs)

  def get_metadata(self):
    return self._metadata

  def get_device(self) -> th.device:
    return self._device

  def get_input_dtype(self, index : int) -> th.dtype:
    np_dtype = self._input_dtypes[index]
    return get_torch_dtype(np_dtype)

  def __call__(self, *args : th.tensor) -> Tuple[th.tensor, ...]:
    return self.forward(*args)



import requests
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
import random
import string

BASE_API_URL = "https://ratio1-debug.ngrok.dev"
API_URL = f"{BASE_API_URL}/constant_endpoint"
CONCURRENT_REQUESTS = 10000  # adjust based on your machine/API
MAX_WORKERS = 200           # max threads at a time

def make_request(i, random_endpoint=False):
  try:
    current_url = API_URL
    if random_endpoint:
      random_endpoint_name = ''.join(random.choices(string.ascii_letters + string.digits, k=10))
      current_url = f"{BASE_API_URL}/{random_endpoint_name}"
    response = requests.get(current_url)
    return f"Request {i}: {response.status_code}"
  except Exception as e:
    return f"Request {i} failed: {e}"

def run_stress_test():
  start_time = time.time()
  with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
    futures = [executor.submit(make_request, i) for i in range(CONCURRENT_REQUESTS)]
    for future in as_completed(futures):
      print(future.result())
  duration = time.time() - start_time
  print(f"\nCompleted {CONCURRENT_REQUESTS} requests in {duration:.2f} seconds")


if __name__ == "__main__":
  run_stress_test()



import os
import sys

from datetime import datetime as dt
from io import BytesIO, TextIOWrapper

import pandas as pd


class _DataFrameMixin(object):
  """
  Mixin for dataframe functionalities that are attached to `libraries.libraries.libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_DataFrameMixin, self).__init__()
    return

  def load_dataframe(
    self, 
    fn, 
    timestamps=None, 
    folder='data', 
    decompress=False, 
    subfolder_path=None
  ):
    """
    if fn ends in ".zip" then the loading will also uncompress in-memory
    """
    import pandas as pd

    assert folder in [None, 'data', 'output', 'models']
    lfld = self.get_target_folder(target=folder)
    
    as_parquet = False
    ext = os.path.splitext(fn)[-1]
    if ext == '.parquet':
      as_parquet = True
    elif decompress:
      ext = '.zip'
    else:
      ext = '.csv'

    if not fn.endswith(ext):
      fn += ext

    if folder is not None:
      datafolder = lfld
      if subfolder_path is not None:
        datafolder = os.path.join(datafolder, subfolder_path.lstrip('/'))
        self.verbose_log("Loading dataframe '{}' from '{}'/'{}'".format(fn, folder, subfolder_path))
      else:
        self.verbose_log("Loading dataframe '{}' from '{}'".format(fn, folder))
      datafile = os.path.join(datafolder, fn)
    else:
      datafile = fn
      self.verbose_log("Loading dataframe '{}'".format(fn))
    #endif

    ext = os.path.splitext(datafile)[-1] # again ?
    file_path = datafile

    self.P("  Loading '{}'...".format(datafile))

    if not os.path.exists(file_path):
      self.P("  Dataframe not found.")
      return

    if as_parquet:
      df = pd.read_parquet(file_path)
    elif ext.lower() == '.zip' or decompress:
      df = pd.read_pickle(file_path)
    else:
      if timestamps is not None:
        if type(timestamps) is str:
          timestamps = [timestamps]
        df = pd.read_csv(file_path, parse_dates=timestamps)
      else:
        df = pd.read_csv(file_path)
    #endif
    return df

  def load_output_dataframe(self,
                            fn, timestamps=None):
    """
    if fn ends in ".zip" then the loading will also uncompress in-memory
    """
    import pandas as pd

    ext = os.path.splitext(fn)[-1]
    file_path = os.path.join(self._outp_dir, fn)
    self.P("Loading '{}'...".format(file_path))
    if ext.lower() == '.zip':
      df = pd.read_pickle(file_path)
    else:
      if timestamps is not None:
        if type(timestamps) is str:
          timestamps = [timestamps]
        df = pd.read_csv(file_path, parse_dates=timestamps)
      else:
        df = pd.read_csv(file_path)
    return df

  def load_abs_dataframe(self,
                         fn, timestamps=None):
    """
    if fn ends in ".zip" then the loading will also uncompress in-memory
    """
    import pandas as pd

    ext = os.path.splitext(fn)[-1]
    file_path = fn
    self.P("Loading '{}'...".format(file_path))
    if ext.lower() == '.zip':
      df = pd.read_pickle(file_path)
    else:
      if timestamps is not None:
        if type(timestamps) is str:
          timestamps = [timestamps]
        df = pd.read_csv(file_path, parse_dates=timestamps)
      else:
        df = pd.read_csv(file_path)
    return df

  def save_dataframe(self,
                     df, fn='', show_prefix=False,
                     folder='data',
                     ignore_index=True, compress=False,
                     mode='w', header=True,
                     also_markdown=False,
                     to_data=None,
                     full_path=None,
                     subfolder_path=None,
                     verbose=True,
                     as_parquet=False,
                     ):
    """
     df: dataframe
     
     fn: name of file
     
     folder: None - absolute path, 'data' - save to data ... etc
     
     show_prefix: add timestamp prefix

     compress: save to zipped pickle

     mode: the writing mode in csv (default 'w' - write). Could be also 'a' - append

     header: bool or list of str, default True
        Write out the column names. If a list of strings is given it is assumed to be aliases for the column names.
        This may be set to False for 'append' mode, for all but not the first save call.

     subfolder_path : str, optional
      A path relative to '_data' or `folder` value (if `to_data=False`) where the dataframe is saved
      Default is None.
      
    as_parquet: bool, default False
      If True then the dataframe is saved as parquet file instead of csv

     (obsolete) to_data: False to save in output dir instead of data dir
     (obsolete) full_path : if full path is specified then file is saved to fn ignoring anything else
    """
    assert isinstance(df, pd.DataFrame)
    
    if to_data is not None:
      self.P("WARNING: `to_data` is obsolete, please use `folder='data'`")
    if full_path is not None:
      self.P("WARNING: `full_path` is obsolete, please use `folder=None`")

    if as_parquet:
      ext = '.parquet'
      save_type = 'Apache Parquet'    
    elif compress:
      ext = '.zip'
      save_type = 'compressed pickle df'
    else:
      ext = '.csv'
      save_type = 'CSV'

    if fn[-4:] != ext:
      fn += ext

    assert folder in [None, 'data', 'output', 'models']
    lfld = self.get_target_folder(target=folder)
    if to_data:
      lfld = self._data_dir

    if subfolder_path is not None:
      lfld = os.path.join(lfld, subfolder_path.lstrip('/'))
      if not os.path.exists(lfld):
        os.makedirs(lfld)

    if lfld is not None:
      file_prefix = '' if not show_prefix else self.file_prefix + "_"
      save_path = lfld
      file_name = file_prefix + fn
      out_file = os.path.join(save_path, file_name)
    else:
      out_file = fn
      save_path, file_name = os.path.split(out_file)

    os.makedirs(os.path.split(out_file)[0], exist_ok=True)

    if verbose:
      self.P("Saving {} (mode='{}') {:<20} [{}] ..{}".format(
        save_type,
        mode, file_name, df.shape, save_path[-30:])
      )
    #endif

    if compress:
      df.to_pickle(out_file)
    elif as_parquet:
      df.to_parquet(out_file)
    else:
      df.to_csv(out_file, index=not ignore_index, mode=mode, header=header)

    if also_markdown:
      has_tabulate = True
      try:
        import imp
        imp.find_module('tabulate')
      except:
        has_tabulate = False
      if not has_tabulate:
        self.raise_error(
          "In order to generate markdown (`also_markdown=True`) you need to install (pip or conda) tabulate")
      fn_md = os.path.join(save_path, file_name + '.md')
      with open(fn_md, 'wt') as fmd:
        fmd.write(df.to_markdown())

    return file_name, out_file

  def update_dataframe(
    self, 
    fn, 
    delta_df, 
    subfolder_path, 
    output_folder='data', 
    compress=False, 
    force_update=False,
    as_parquet=False,
  ):
    assert output_folder in ['data', 'output']
    datafile = self.get_file_path(
      fn=fn,
      folder=output_folder,
      subfolder_path=subfolder_path,
      force=True
    )
    if datafile is None:
      self.P("update_dataframe_from_data failed due to missing {}".format(datafile), color='error')
      return False

    with self.managed_lock_resource(datafile):
      result = None
      try:
        data = self.load_dataframe(
          fn=fn,
          subfolder_path=subfolder_path,
          timestamps=None,
          folder=output_folder,
          decompress=compress
        )

        if data is not None or force_update:
          if data is None and force_update:
            data = delta_df
          else:
            data = pd.concat([data, delta_df])

          self.save_dataframe(
            df=data, fn=fn,
            folder=output_folder,
            ignore_index=True, compress=compress,
            mode='w', header=True,
            subfolder_path=subfolder_path,
            verbose=True,
            as_parquet=as_parquet,
          )
          result = True
      except Exception as e:
        self.P("update_pickle_from_data failed: {}".format(e), color='error')
        result = False
    # endwith lock
    return result

  def update_dataframe_from_output(self, fn, delta_df, subfolder_path, compress=False, force_update=False):
    return self.update_dataframe(
      fn=fn,
      delta_df=delta_df,
      subfolder_path=subfolder_path,
      output_folder='output',
      compress=compress,
      force_update=force_update
    )

  def update_dataframe_from_data(self, fn, delta_df, subfolder_path, compress=False, force_update=False):
    return self.update_dataframe(
      fn=fn,
      delta_df=delta_df,
      subfolder_path=subfolder_path,
      output_folder='data',
      compress=compress,
      force_update=force_update
    )

  def save_dataframe_current_time(self,
                                  df, fn=''):
    """
    saves a DataFrame in 'output' folder with current time prefix
    """
    file_prefix = dt.now().strftime("%Y%m%d_%H%M%S")
    csvfile = os.path.join(self._outp_dir, file_prefix + fn + '.csv')
    df.to_csv(csvfile)
    return csvfile

  @staticmethod
  def get_dataframe_info(df):
    # setup the environment
    old_stdout = sys.stdout
    sys.stdout = TextIOWrapper(BytesIO(), sys.stdout.encoding)
    # write to stdout or stdout.buffer
    df.info()
    # get output
    sys.stdout.seek(0)  # jump to the start
    out = sys.stdout.read()  # read output
    # restore stdout
    sys.stdout.close()
    sys.stdout = old_stdout

    str_result = "DataFrame info:\n" + out
    return str_result

  def save_dataframe_to_hdf(self,
                            df, h5_file, h5_format='table'):
    """
     saves pandas df to h5_file in _data folder
     assume h5_file DOES NOT HAVE path info
    """
    assert "/" not in h5_file
    assert "\\" not in h5_file
    table_name, ext = os.path.splitext(h5_file)
    out_file = os.path.join(self._data_dir, h5_file)
    self.verbose_log("Saving ...{}".format(out_file[-40:]))
    df.to_hdf(out_file, key='table_' + table_name,
              append=False, format=h5_format)
    self.verbose_log("Done saving ...{}".format(out_file[-40:]), show_time=True)
    return

  def load_dataframe_from_hdf(self,
                              h5_file):
    """
     loads pandas dataframe from h5 file store
     assume h5_file DOES NOT HAVE path info
    """
    import pandas as pd

    assert "/" not in h5_file
    assert "\\" not in h5_file
    table_name, ext = os.path.splitext(h5_file)
    out_file = os.path.join(self._data_dir, h5_file)
    self.verbose_log("Loading ...{}".format(out_file[-40:]))
    df = pd.read_hdf(out_file, key='table_' + table_name)
    self.verbose_log("Done loading ...{}".format(out_file[-40:]), show_time=True)
    return df

  @staticmethod
  def drop_constant_columns(df):
    """
    Drops constant value columns of pandas dataframe.
    """
    return df.loc[:, (df != df.iloc[0]).any()]

  @staticmethod
  def remove_constant_columns(df):
    """
    removes constant dataframe columns
    """
    if df.shape[0] <= 1:
      return df
    good_columns = []
    for col in df.columns:
      if df[col].astype(str).nunique() > 1:
        good_columns.append(col)
    return df[good_columns]



import os
import numpy as np
from naeural_core.bc import DefaultBlockEngine, BCct
import json
from naeural_core import Logger

if __name__ == '__main__':
  l = Logger('BC', base_folder='.', app_folder='_local_cache')
  config_send = {
    BCct.K_PEM_LOCATION   : 'data',
    BCct.K_PEM_FILE       : 'e2.pem'  ,
    BCct.K_PASSWORD       : None,
  }

  config_bandit = {
    BCct.K_PEM_LOCATION   : 'data',
    BCct.K_PEM_FILE       : 'e2-bandit.pem'  ,
    BCct.K_PASSWORD       : None,
  }

  config_recv = {
    BCct.K_PEM_LOCATION   : 'data',
    BCct.K_PEM_FILE       : 'e2-recv.pem'  ,
    BCct.K_PASSWORD       : None,
  }
  
  c_sender = DefaultBlockEngine(name='e2-sender', log=l, config=config_send)
  
  c_s1 = DefaultBlockEngine(name='e2-sender', log=l, config=config_send)
    
  data = {str(x):x for x in np.random.randint(0,10, size=10).tolist()}
  data['5'] = {str(x):x for x in np.random.randint(10,20, size=5).tolist()}
  signature = c_sender.sign(data, use_digest=True)
  l.P("Data:\n{}".format(json.dumps(data, indent=4)))
  
  
  data_json = json.dumps(data)

  ###
  ###
  ###

  c_bandit = DefaultBlockEngine(name='e2-bandit', log=l, config=config_bandit)
  intercepted_data = json.loads(data_json)
  intercepted_data['5'] = None
  original_sender = intercepted_data.pop(BCct.SENDER)
  original_sign = intercepted_data.pop(BCct.SIGN)
  original_hash = intercepted_data.pop(BCct.HASH, None)
  use_digest = False
  # if original_hash is not None:
  #   use_digest = True
  fake_signature = c_bandit.sign(intercepted_data, add_data=True, use_digest=use_digest)
  intercepted_data[BCct.SIGN] = fake_signature
  intercepted_data[BCct.SENDER] = original_sender
  data_json_bandit = json.dumps(intercepted_data)
  
  ###
  ###
  ###
  
  c_receiver = DefaultBlockEngine(name='e2-consumer', log=l, config=config_recv)
  
  received_data_bandit = json.loads(data_json_bandit)  
  res = c_receiver.verify(dct_data=received_data_bandit)
  l.P("Received:\n{}\n\nSignature: {}".format(
    json.dumps(json.loads(data_json_bandit), indent=4),
    res
  ))
  
  received_data_original = json.loads(data_json)  
  res = c_receiver.verify(dct_data=received_data_original)
  l.P("Received:\n{}\n\nSignature: {}".format(
    json.dumps(json.loads(data_json), indent=4),
    res
  )) 
  
  
  


from naeural_core import Logger

if __name__ == '__main__':
  
  STAGE = 4
  
  l = Logger('ZIP', base_folder='.', app_folder='_local_cache')
  
  logs = l.get_log_files()
  
  if STAGE == 1:
    l.add_file_to_zip('core/xperimental/zip/test1.zip', logs[0])
    
  
  if STAGE == 2:
    l.add_files_to_zip('core/xperimental/zip/test1.zip', logs[1:3])
  
  
  if STAGE == 3:
    l.add_files_to_zip('core/xperimental/zip/test2.zip', logs[:5])


  if STAGE == 4:
    l.add_files_to_zip('core/xperimental/zip/test2.zip', logs[5:10])



"""
TODO:


  - Solve the issue for set-contention in the chain storage when two nodes try to set the same key at the same time
    - (better) implement a lock mechanism for the chain storage when setting a key value that 
      will allow multiple nodes to compete for the set-operation and thus load-balance it
      OR
    - implement set-value moving TOKEN via the network
  

  - peer-to-peer managed:
    - node 1 registers a app_id X via a list of peers that includes node 1, node 2 and node 3
    - node 2 sets value by:
      - key
      - value 
      - app_id X
    - node 2 broadcasts the set opration to all app_id X peers (not all peers)
    - node 2 waits for confirmations from at least half of the app_id peers
    - if node 4 tries to set key in app_id X, it will be rejected
    
  


"""


from naeural_core.business.base.network_processor import NetworkProcessorPlugin 

_CONFIG = {
  **NetworkProcessorPlugin.CONFIG,
  
  'ALLOW_EMPTY_INPUTS' : True,  
  "ACCEPT_SELF" : False,
  
  "FULL_DEBUG_PAYLOADS" : False,
  "CHAIN_STORE_DEBUG" : False, # main debug flag
  
  
  "MIN_CONFIRMATIONS" : 1,
  
  "CHAIN_PEERS_REFRESH_INTERVAL" : 60,

  'VALIDATION_RULES' : { 
    **NetworkProcessorPlugin.CONFIG['VALIDATION_RULES'],
  },  
}

__VER__ = '0.8.1'

class ChainStoreBasePlugin(NetworkProcessorPlugin):
  CONFIG = _CONFIG
  
  CS_STORE = "SSTORE"
  CS_CONFIRM = "SCONFIRM"
  CS_DATA = "CHAIN_STORE_DATA"
  CS_PEERS = "PEERS"

  CS_CONFIRM_BY = "confirm_by"
  CS_CONFIRM_BY_ADDR = "confirm_by_addr"
  CS_CONFIRMATIONS = "confirms"
  CS_MIN_CONFIRMATIONS = "min_confirms"
  CS_OP = "op"
  CS_KEY = "key"
  CS_VALUE = "value"
  CS_OWNER = "owner"
  CS_READONLY = "readonly"
  CS_TOKEN = "token"
  CS_STORAGE_MEM = "__chain_storage" # shared memory key
  CS_GETTER = "__chain_storage_get"
  CS_SETTER = "__chain_storage_set"
  
  
  
  
  def on_init(self):
    super().on_init() # not mandatory anymore?
    
    self.P(" === ChainStoreBasicPlugin INIT")
    
    self.__chainstore_identity = "CS_MSG_{}".format(self.uuid(7))
    
    self.__ops = self.deque()
    
    try:
      self.__chain_storage = self.cacheapi_load_pickle(default={}, verbose=True)
    except Exception as e:
      self.P(f" === Chain storage could not be loaded: {e}")
      self.__chain_storage = {}
      
    memory = self.plugins_shmem
      

    ## DEBUG ONLY:
    if self.CS_STORAGE_MEM in memory:
      self.P(" === Chain storage already exists", color="r")
      self.__chain_storage = memory[self.CS_STORAGE_MEM]
    ## END DEBUG ONLY
    
    memory[self.CS_STORAGE_MEM] = self.__chain_storage
    memory[self.CS_GETTER] = self._get_value
    memory[self.CS_SETTER] = self._set_value
    
    self.__last_chain_peers_refresh = 0
    self.__chain_peers = []
    self.__maybe_refresh_chain_peers()
    return
  
  
  def __debug_dump_chain_storage(self):
    if self.cfg_chain_store_debug:
      self.P(" === Chain storage dump:\n{}".format(self.json_dumps(self.__chain_storage, indent=2)))
    return
   
  
  
  
  def __maybe_refresh_chain_peers(self):
    """
    This method refreshes the chain peers list from the network using the whitelist generated
    by the blockchain engine. This means it will allow broadcasting the local keys to all the
    oracle nodes in the network as well as the manually added nodes. This pretty much covers 
    the "private" part of the ChainStore.
    
    However the chain storage should be also accessible to all the nodes in the network so that 
    they can ALL the values stored in the chain storage publicly
        
    """
    if (self.time() - self.__last_chain_peers_refresh) > self.cfg_chain_peers_refresh_interval:
      _chain_peers = self.bc.get_whitelist(with_prefix=True)
      # now check and preserve only online peers
      self.__chain_peers = [
        peer for peer in _chain_peers if self.netmon.network_node_is_online(peer)
      ]
      self.__last_chain_peers_refresh = self.time()
    return
  
  
  def __send_data_to_chain_peers(self, data, peers=None):
    # check if list or str
    send_to = self.deepcopy(self.__chain_peers)
    if isinstance(peers, (str, list)) and len(peers) > 0:
      if isinstance(peers, str):
        peers = [peers]
      # end if not a list
      peers = [peer for peer in peers if peer not in self.__chain_peers]
      send_to.extend(peers)
    # end if peers
      
    self.send_encrypted_payload(node_addr=send_to, **data)
    return
  
  
  def __get_min_peer_confirmations(self):
    if self.cfg_min_confirmations is not None and self.cfg_min_confirmations > 0:
      return self.cfg_min_confirmations    
    return len(self.__chain_peers) // 2 + 1
  
  
  def __save_chain_storage(self):
    self.cacheapi_save_pickle(self.__chain_storage, verbose=True)
    self.__last_chain_storage_save = self.time()
    return
  
  ## START setter-getter methods

  def __get_key_value(self, key):
    return self.__chain_storage.get(key, {}).get(self.CS_VALUE, None)


  def __get_key_owner(self, key):
    return self.__chain_storage.get(key, {}).get(self.CS_OWNER, None)
  
  
  def __get_key_readonly(self, key):
    return self.__chain_storage.get(key, {}).get(self.CS_READONLY, False)


  def __get_key_token(self, key):
    return self.__chain_storage.get(key, {}).get(self.CS_TOKEN, None)


  def __get_key_confirmations(self, key):
    return self.__chain_storage.get(key, {}).get(self.CS_CONFIRMATIONS, 0)

  
  def __get_key_min_confirmations(self, key):
    return self.__chain_storage.get(key, {}).get(self.CS_MIN_CONFIRMATIONS, 0)


  def __reset_confirmations(self, key):
    self.__chain_storage[key][self.CS_CONFIRMATIONS] = 0
    self.__chain_storage[key][self.CS_MIN_CONFIRMATIONS] = self.__get_min_peer_confirmations()
    return


  def __increment_confirmations(self, key):
    self.__chain_storage[key][self.CS_CONFIRMATIONS] += 1
    return
  
  def __set_confirmations(self, key, confirmations):
    self.__chain_storage[key][self.CS_CONFIRMATIONS] = confirmations
    return


  def __set_key_value(self, key, value, owner,  readonly=False, token=None, local_sync_storage_op=False):
    """
    This method is called to set a key-value pair in the chain storage.
    
    Parameters:
    ----------
    
    key : str
      The key to set the value for
    
    value : any
      The value to set
      
    owner : str
      The owner of the key-value pair
      
    readonly : bool
      If True the key-value pair will be readonly and cannot be overwritten by other owners
      
    token: any
      A token to be used for the set operation. If the token is not None, any read/write operations
      will have to have the same token.
      
    local_sync_storage_op : bool
      If `True` will only set the local kv pair without broadcasting to the network. 
      This operation is used for remote sync when a node receives a set operation from 
      the network and needs to set the value in the local chain storage replica.
      
    """
    # key should be composed of the chainstore app identity and the actual key
    # so if two chainstore apps are running on the same node, they will not overwrite each other
    # also this way we can implement chaistore app allow-listing
    chain_key = key
    self.__chain_storage[chain_key] = {
      self.CS_KEY       : key,
      self.CS_VALUE     : value,
      self.CS_OWNER     : owner,
      self.CS_READONLY  : readonly,
      self.CS_TOKEN     : token,
    }    
    self.__reset_confirmations(key)
    if local_sync_storage_op:
      # set the confirmations to -1 to indicate that the key is remote synced on this node
      self.__set_confirmations(key, -1) # set to -1 to indicate that the key is remote synced on this node
    self.__save_chain_storage()
    return


  def _set_value(
    self, 
    key, 
    value, 
    owner=None, 
    readonly=False,
    token=None,
    local_sync_storage_op=False, 
    peers=None,
    debug=False, 
  ):
    """ 
    This method is called to set a value in the chain storage.
    If called locally will push a broadcast request to the network, 
    while if called from the network will set the value in the chain storage.
    
    Parameters:
    ----------
    
    key : str
      The key to set the value for
      
    value : any
      The value to set
      
    owner : str 
      The owner of the key-value pair
      
    readonly : bool
      If True the key-value pair will be readonly and cannot be overwritten by other owners
      
    token: any
      A token to be used for the set operation. If the token is not None, any read/write operations 
      will have to have the same token.
            
    local_sync_storage_op : bool
      If True will only set the local kv pair without broadcasting to the network
      
    peers : list
      A list of peers to send the data to. If None, will use only the chain peers list.

    debug : bool
      If True will print debug messages
      
      
    Returns:
    --------
    
    
    
    """
    where = "FROM_LOCAL: " if not local_sync_storage_op else "FROM_REMOTE: "
    debug = debug or self.cfg_chain_store_debug
    debug_val = str(value)[:20] + "..." if len(str(value)) > 20 else str(value)
    if owner is None:
      owner = self.get_instance_path()
    need_store = True
    existing_owner = None
    existing_value = None
    if key in self.__chain_storage:
      existing_value = self.__get_key_value(key)
      existing_owner = self.__get_key_owner(key)
      is_readonly = self.__get_key_readonly(key)
      existing_token = self.__get_key_token(key)
      if token != existing_token:
        if debug:
          self.P(f" === Key {key} has a different token {existing_token} from {existing_owner} than the one provided {token} from {owner}", color='r')
        need_store = False
      elif existing_value == value:
        if debug:
          self.P(f" === Key {key} stored by {existing_owner} has the same value")
        need_store = False
      elif is_readonly and existing_owner != owner:
        if debug:
          self.P(f" === Key {key} readonly by {existing_owner} (requester: {owner})", color='r')
        need_store = False
    # end if key in chain storage
    if need_store:
      if debug:
        set_or_overwrite = "overwriting" if existing_owner not in [None, owner] else "setting"        
        self.P(f" === {where}{set_or_overwrite} <{key}> = <{debug_val}> by {owner} (orig: {existing_owner}), is_remote={local_sync_storage_op}")
      self.__set_key_value(
        key=key, value=value, owner=owner, 
        local_sync_storage_op=local_sync_storage_op, 
        readonly=readonly, token=token,
      )
      if not local_sync_storage_op:      
        # now send set-value (including confirmation request) to all
        op = {      
            self.CS_OP        : self.CS_STORE,
            self.CS_KEY       : key,        
            self.CS_VALUE     : value,   
            self.CS_OWNER     : owner,
            self.CS_TOKEN     : token,
            self.CS_READONLY  : readonly, # if the key is readonly, it will not be overwritten by other owners
            self.CS_PEERS     : peers,
        }
        self.__ops.append(op)
        if debug:
          self.P(f" === {where} key {key} locally stored for {owner}. Now waiting for confirmations...")
        # at this point we can wait until we have enough confirmations
        _timeout = self.time() + 10
        _done = False
        _prev_confirm = 0
        _max_retries = 2
        _retries = 0
        while not _done: # this LOCKS the calling thread set_value
          recv_confirm = self.__get_key_confirmations(key)
          if recv_confirm > _prev_confirm:
            _prev_confirm = recv_confirm
            if debug:
              self.P(f" === {where}Key received '{key}' has {recv_confirm} confirmations")
          if recv_confirm >= self.__get_key_min_confirmations(key):
            if debug:
              self.P(f" === {where}KEY CONFIRMED '{key}': has enough ({recv_confirm}) confirmations")
            _done = True
            need_store = True
            continue
          elif self.time() > _timeout:
            if debug:
              self.P(f" === {where}Key '{key}' has not enough confirmations after timeout", color='r')
            _retries += 1
            if _retries > _max_retries:
              if debug:
                self.P(f" === {where}Key '{key}' has not enough confirmations after {_max_retries} retries", color='r')
              _done = True
              need_store = False
            else:
              if debug:
                self.P(f" === {where}Retrying key '{key}' with timeout...", color='r')
              self.__ops.append(op)
              _timeout = self.time() + 10
            # end if retries
          # end if timeout
          self.sleep(0.100)  # sleep for 100ms to give protocol sync time
        # end while not done
      else:
        if debug:
          self.P(f" === {where}{key} locally sync-stored for remote {owner}")
      # end if not sync_storage
    # end if need_store
    return need_store


  def _get_value(self, key, token=None, get_owner=False, debug=False):
    """ This method is called to get a value from the chain storage """
    debug = debug or self.cfg_chain_store_debug
    if debug:
      self.P(f" === Getting value for key {key}")

    existing_token = self.__get_key_token(key)
    result_value, result_owner = None, None
    if token != existing_token:
      if debug:
        self.P(f" === Key {key} has a different token {existing_token} than the one provided {token}", color='r')
    else:
      result_value = self.__get_key_value(key)
      if get_owner:
        result_owner = self.__get_key_owner(key)
    # end if token
    if get_owner:
      return result_value, result_owner
    return result_value
  
  ### END setter-getter methods


  def __maybe_broadcast(self):
    """ 
    This method is called to broadcast the chain store operations to the network.
    For each operation in the queue, a broadcast is sent to the network    
    """
    if self.cfg_chain_store_debug and len(self.__ops) > 0:
      self.P(f" === Broadcasting {len(self.__ops)} chain store {self.CS_STORE} ops to {self.__chain_peers}")
    while len(self.__ops) > 0:
      data = self.__ops.popleft()
      peers = data.get(self.CS_PEERS, None)
      payload_data = {
        self.CS_DATA : data
      }
      self.__send_data_to_chain_peers(payload_data, peers=peers)
    return


  def __exec_store(self, data, peers=None):
    """ 
    This method is called when a store operation is received from the network. The method will:
      - set the value in the chain storage
      - send a ecrypted confirmation of the storage operation to the network
    """
    key = data.get(self.CS_KEY, None)
    value = data.get(self.CS_VALUE , None)
    owner = data.get(self.CS_OWNER, None)
    readonly = data.get(self.CS_READONLY, False) # if the key is readonly local node consumers cannot overwrite it
    token = data.get(self.CS_TOKEN, None)
    if self.cfg_chain_store_debug:
      self.P(f" === REMOTE: Exec remote-to-local-sync store for {key}={value} by {owner}")
    result = self._set_value(
      key, value, owner=owner, 
      token=token, readonly=readonly,
      local_sync_storage_op=True,
    )
    if result:
      # now send confirmation of the storage execution
      if self.cfg_chain_store_debug:
        self.P(f" === REMOTE: {self.CS_CONFIRM} for {key} of {owner} to {self.__chain_peers}")
      data = {
        self.CS_DATA : {
          self.CS_OP : self.CS_CONFIRM,
          self.CS_KEY: key,
          self.CS_VALUE : value,
          self.CS_OWNER : owner,
          self.CS_CONFIRM_BY : self.get_instance_path(),
          self.CS_CONFIRM_BY_ADDR : self.ee_addr,
        }
      }
      self.__send_data_to_chain_peers(data, peers=peers)
    else:
      if self.cfg_chain_store_debug:
        self.P(f" === REMOTE: Store for {key}={value} of {owner} failed", color='r')
    return


  def __exec_received_confirm(self, data):
    """ This method is called when a confirmation of a broadcasted store operation is received from the network """
    key = data.get(self.CS_KEY, None)
    value = data.get(self.CS_VALUE, None)
    owner = data.get(self.CS_OWNER, None)
    confirm_by = data.get(self.CS_CONFIRM_BY, None)
    op = data.get(self.CS_OP, None)
    
    local_owner = self.__get_key_owner(key)
    local_value = self.__get_key_value(key)
    if self.cfg_chain_store_debug:
      self.P(f" === LOCAL: Received {op} from {confirm_by} for  {key}={value}, owner{owner}")
    if owner == local_owner and value == local_value:
      self.__increment_confirmations(key)
      if self.cfg_chain_store_debug:
        self.P(f" === LOCAL: Key {key} confirmed by {confirm_by}")
    return

  @NetworkProcessorPlugin.payload_handler()
  def default_handler(self, payload):
    sender = payload.get(self.const.PAYLOAD_DATA.EE_SENDER, None)
    alias = payload.get(self.const.PAYLOAD_DATA.EE_ID, None)
    destination = payload.get(self.const.PAYLOAD_DATA.EE_DESTINATION, None)
    is_encrypted = payload.get(self.const.PAYLOAD_DATA.EE_IS_ENCRYPTED, False)
    destination = destination if isinstance(destination, list) else [destination]
    decrypted_data = self.receive_and_decrypt_payload(data=payload)    
    # DEBUG AREA
    if self.cfg_chain_store_debug:
      from_myself = sender == self.ee_addr
      str_sender = sender if not from_myself else f"{sender} (myself)"
      self.P(f" === PAYLOAD_CSTORE: from {str_sender} (enc={is_encrypted})")
      if self.ee_addr in destination:
        self.P(f" === PAYLOAD_CSTORE: received for me")
      else:
        if not from_myself:
          self.P(f" === PAYLOAD_CSTORE: to {destination} (not for me {self.ee_addr})", color='r')
        return
    # try to decrypt the payload
    if self.cfg_chain_store_debug:
      if decrypted_data is None or len(decrypted_data) == 0:
        self.P(f" === PAYLOAD_CSTORE: FAILED decrypting payload", color='r')
      else:
        self.P(f" === PAYLOAD_CSTORE: decrypted payload OK")
    # END DEBUG AREA    
    # get the data and call the appropriate operation method
    data = decrypted_data.get(self.CS_DATA, {})
    operation = data.get(self.CS_OP, None)
    owner = data.get(self.CS_OWNER, None)
    if self.cfg_chain_store_debug:
      if operation is None:
        self.P(f" === PAYLOAD_CSTORE: NO OPERATION from data: {data}", color='r')
      else:
        self.P(f" === PAYLOAD_CSTORE: {operation=} from {alias=} {owner=}")
    if operation == self.CS_STORE:
      self.__exec_store(data, peers=sender) # make sure you send also to the sender
    elif operation == self.CS_CONFIRM:
      self.__exec_received_confirm(data)
    return
  

  
  def process(self):
    self.__maybe_refresh_chain_peers()
    self.__maybe_broadcast()
    return 



from time import time
import functools

def stopwatch_decorator(func):
  @functools.wraps(func)
  def wrapper(*arg, **kwargs):
    start = time()
    res = func(*arg, **kwargs)
    return res, time()-start

  return wrapper



import torch as th
import os

class ONNXMixin:
  def _has_valid_onnx_precision(self, config : dict) -> bool:
    """
    Checks if the precision for this model is compatible
    with ONNX serving.

    Parameters
    ----------
    config: dict, the config of the model

    Returns
    -------
    bool - True if the model precision allows us to run with the
      ONNX serving
    """
    precision = config.get('precision').lower()
    if self.cfg_fp16 and precision is not None and precision != "fp16":
      return False
    return True

  def _prepare_onnx_model(self,
    url : str = None,
    fn_model : str = None,
    post_process_classes : bool = False,
    return_config : bool = False,
    **kwargs
  ):
    """
    Creates an ONNX backend model, possibliy downloading it.

    Parameters
    ---------
    url: str
      The URL where we can download the ONNX model from. If None, the
      URL value will be picked as the default URL of the serving (specified
      by ONNX_URL)

    fn_model : str
      The filename of the ONNX model. If None, the default ONNX filename value
      will be used (specified by MODEL_ONNX_URL)

    post_process_classes : bool
      If True updates the returned configuration, changing keys from strings
      to integers

    return_config : bool
      If True, returns a tuple of (model, model configuration, filename)
      If False, only returns the model.

    Returns
    -------
    if return_config is False
      res - the ONNXModel that was loaded

    if return_config is True
      res - tuple of (ONNXModel, config) where the ONNXModel is the model that
        was loaded/prepared, config is the configuration of the model (dict)
    """
    from naeural_core.serving.base.backends.onnx import ONNXModel
    if 'batch_size' not in kwargs.keys():
      self.P("batch size not passed as parameter when preparing model")
      raise ValueError("Batch size required for ONNX model")
    #endif check for batch size
    max_batch_size = kwargs['batch_size']

    self.P("Preparing {} ONNX model {}...".format(self.server_name, self.version))
    if url is None:
      url = self.cfg_onnx_url
    if fn_model is None:
      fn_model = self.cfg_model_onnx_filename

    if fn_model is None:
      self.P('ONNX model not specified')
      # No ONNX file was specified in the configuration so bail out.
      raise ValueError("ONNX model not available")
    #endif check model path

    if not self._th_device_eq(self.dev, th.device('cpu')):
      self.P("Incompatible device for ONNX backend: {}".format(self.dev))
      raise ValueError("Trying to use ONNX backend with incompatible device")
    #endif check torch device type

    model_dir = self.download_model_for_backend(url=url, fn_model=fn_model, backend='onnx')
    fn_path = os.path.join(model_dir, fn_model)

    if fn_path is not None:
      self.P("Using ONNX model {} ({:.03f} MB) at `{}` using map_location: {} on python v{}...".format(
          fn_model,
          self.os_path.getsize(fn_path) / 1024 / 1024,
          fn_path,
          self.dev,
          self.python_version()
        ),
        color='y'
      )
    # Just load or rebuild the model.
    model = ONNXModel()
    self.P("Trying to load from {}".format(fn_path))
    model.load_model(fn_path, max_batch_size, self.dev, self.cfg_fp16)
    config = model.get_metadata()

    err_keys = ['torch']
    env_versions = {
      'python': self.python_version(),
      'torch': self.th.__version__,
      'torchvision': self.tv.__version__
    }
    self.check_versions(config, fn_path, env_versions, err_keys)
    if post_process_classes:
      config = self._process_config_classes(config)

    self.P("  Model config:\n{}".format(self.log.dict_pretty_format(config)))
    return (model, config) if return_config else model



import json
import numpy as np
from datetime import datetime

class AbstractSensor:
  def __init__(self, **kwargs):
    self._pos = 0
    self._setup_connection(**kwargs)
    return
  
  def _setup_connection(self, **kwargs):
    raise NotImplementedError()
  
  def _get_single_observation(self):
    raise NotImplementedError()
    
  def _has_training_data(self):
    raise NotImplementedError()
    
  def _get_training_data(self):
    raise NotImplementedError()
    
  def release(self):
    return
  
  def has_training_data(self):
    has_data = self._has_training_data()
    return has_data

  def maybe_reconnect(self):
    return True
  
  def get_observation(self, as_json=False):
    data = self._get_single_observation()
    if as_json:
      if isinstance(data, np.ndarray):
        data = data.tolist()
      elif isinstance(data, dict):
        for k in data:
          if isinstance(data[k], np.ndarray):
            data[k] = data[k].tolist()
          elif isinstance(data[k], datetime):
            data[k] = str(data[k])            
      data = json.dumps(data, indent=4)
    return data
  
  def get_train_data(self):
    if self._has_training_data():
      data = self._get_training_data()
      return data
    return None
  
  def get_training_data(self):
    return self.get_train_data()
  
  
  def get_datastream_len(self):
    return vars(self).get('_datastream_len', None)
  
  def get_position(self):
    return self._pos
   
  
  def __repr__(self):
    str_res = "{}:\n{}".format(
      self.__class__.__name__,
      self.get_observation(as_json=True)
      )
    return str_res


from naeural_core.local_libraries import _ConfigHandlerMixin
from naeural_core import DecentrAIObject
from naeural_core import Logger


_CONFIG1 = {
  'P1' : 1,
  'TEST_PRC' : 1.0,
  
  'P_2_1' : 3,
  
  'P_A' : 't1',

  'D_A' : {},

  'L_A' : ['test'],
  
  'VALIDATION_RULES' : {
    'P1' : {
      'TYPE' : 'int',
      'MIN_VAL' : 1,
      'MAX_VAL' : 1,
      },
    
    'P_A' : {
      'TYPE' : 'str',
      'MIN_LEN' : 2,
      'EXCLUDED_LIST' : ['XXXX'],
      'ACCEPTED_VALUES' : ['t1','t2']
      },
    
    'TEST_PRC' :{
      'TYPE' : 'float',
      'MAX_VAL' : 1,
      'MIN_VAL' : 0,
      },
    'D_A' :{
      'TYPE' : 'dict',
      },
    'L_A' :{
      'TYPE' : 'list',
      },
    
    # 'TEST1' : { # this key cannot be found in validation
    #   'TYPE' : 'int'
    #   }
    }
  
}

class Test(DecentrAIObject, _ConfigHandlerMixin):
  CONFIG = _CONFIG1
  def __init__(self, config, **kwargs):
    self.config = config
    super().__init__(**kwargs)    
    return
    
  def startup(self):
    super().startup()    
    self.config = self._merge_prepare_config(
      default_config=self.config,
      
      debug=False
    )
    self.config_data = self.config
    self.create_config_handlers()
    self.validate()
    
      
  def update(self, cfg):
    self.config = self._merge_prepare_config(
      default_config=self.config,
      debug=False
    )      
    return

if __name__ == '__main__':
  l = Logger('CFG', base_folder='.', app_folder='_local_cache')
  
  
  eng = Test(config=_CONFIG1, log=l)
  res, msgs = eng.run_validation_rules(verbose=100)
  l.P("Result: {}, {}".format(res, msgs), color=None if res else 'r')
  



import torch as th
from naeural_core.local_libraries.nn.th.layers import DenseBlock
from naeural_core.local_libraries.nn.utils import get_dropout


class DenseEncoder(th.nn.Module):
  def __init__(self, input_size, layer_sizes, dropout=('constant', 0.3), use_batch_norm='pre', activation=None):
    super(DenseEncoder, self).__init__()
    if isinstance(dropout, tuple):
      self.dropout_type, self.dropout = dropout
    else:
      self.dropout_type = 'constant'
      self.dropout = dropout
    self.output_size = layer_sizes[-1]
    n_layers = len(layer_sizes)

    layers = []
    current_input = input_size
    for i in range(n_layers):
      current_dropout = get_dropout(
        dropout_type=self.dropout_type,
        dropout=self.dropout,
        step=i,
        max_step=n_layers - 1
      )
      layers.append(
        DenseBlock(
          in_channels=current_input,
          n_layers=layer_sizes[i],
          dropout=current_dropout,
          use_batch_norm=use_batch_norm,
          activation=activation
        )
      )
      current_input = layer_sizes[i]

    self.layers = th.nn.Sequential(*layers)

    return

  def forward(self, th_x):
    return self.layers(th_x)




import paho.mqtt.publish as publish

DEVICE_ID = '1234'

_CONFIG_SENDER = {
  'MQTT_HOST': 'HOSTNAME',
  'MQTT_PORT': 1883,
  'MQTT_USER': 'USERNAME',
  'MQTT_PASS': 'PASSWORD',
  'MQTT_PATH': 'PATH',
  'MQTT_TOPIC': 'config1',
  'MQTT_QOS': 1
  }

_CONFIG = _CONFIG_SENDER

if __name__ == '__main__':  
  topic = '{}/{}'.format(_CONFIG['MQTT_PATH'], _CONFIG['MQTT_TOPIC'])
  
  for i in range(10):
    payload = 'Msg {}'.format(i)
    
    publish.single(
      topic=topic, 
      payload=payload, 
      qos=1,
      hostname=_CONFIG['MQTT_HOST'],
      port=_CONFIG['MQTT_PORT'], 
      client_id=DEVICE_ID, 
      keepalive=60, 
      auth={'username': _CONFIG['MQTT_USER'], 'password': _CONFIG['MQTT_PASS']}
      )
    
  
  


import cv2 as cv
import numpy as np

if __name__ == '__main__':
  FPS = 12
  NR = 300
  path = 'C:/Users/ETA/Dropbox/DATA/_vapor_data/__sources/movies/lpr/SeeTec.avi'
  
  cap = cv.VideoCapture(path)
  fourcc = cv.VideoWriter_fourcc(*'MP4V')
  out = cv.VideoWriter('output.mp4', fourcc, FPS, (1280, 1024))
  crt = 0
  while cap.isOpened():    
    ret, frame = cap.read()
    crt += 1
    print(frame.shape)
    out.write(frame)
    cv.imshow('frame', frame)
    if cv.waitKey(1) == ord('q'):
      break
    
    if crt >= NR:
      break
  #endwhile
  cap.release()
  out.release()
  cv.destroyAllWindows()


import os
import pandas as pd
import numpy as np
import operator
from functools import partial
from scipy import sparse

from naeural_core import DecentrAIObject

from naeural_core.utils.predictive_analytics.timeseries.numpyize_constants import NumpyizeConstants
from naeural_core.utils.predictive_analytics.timeseries import connectors
from naeural_core.utils.predictive_analytics.timeseries.connector_constants import ConnectorConstants
from naeural_core.utils.predictive_analytics.timeseries.binarization import BaseDataBinarization

def tuple_scalar(x):
  if type(x) is not tuple:
    return tuple([x,])
  return x
#enddef

def save_dict_to_hdf5(dct, filename, root='/', write_mode='w'):
  import h5py
  
  if not root.startswith('/'):
    root = '/' + root

  if not root.endswith('/'):
    root = root + '/'

  with h5py.File(filename, write_mode) as h5file:
    recursively_save_dict_contents_to_group(h5file, root, dct)


def recursively_save_dict_contents_to_group(h5file, path, dct):
  base_instances = (
    np.ndarray,
    np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32,
    np.float16, np.float32, np.float64,
    str, bytes, int, float,
  )
  for key, item in dct.items():
    if isinstance(item, base_instances):
      h5file[path + key] = item
    elif isinstance(item, dict):
      recursively_save_dict_contents_to_group(h5file, path + key + '/', item)
    else:
      raise ValueError('Cannot save %s type' % type(item))


def load_dict_from_hdf5(filename, root='/'):
  import h5py
  
  if not root.startswith('/'):
    root = '/' + root

  if not root.endswith('/'):
    root = root + '/'

  with h5py.File(filename, 'r') as h5file:
    return recursively_load_dict_contents_from_group(h5file, root)


def recursively_load_dict_contents_from_group(h5file, path):
  import h5py

  ans = {}
  for key, item in h5file[path].items():
    if isinstance(item, h5py.Dataset):
      ans[key] = item[()]
    elif isinstance(item, h5py.Group):
      ans[key] = recursively_load_dict_contents_from_group(h5file, path + key + '/')
  return ans

def is_void_list(lst):
  if lst is None:
    return True

  if isinstance(lst, list):
    return len(lst) == 0
  else:
    raise ValueError("Object of type `list` expected in `is_void_list`")


class TimeseriesDataMaster(BaseDataBinarization):
  def __init__(self,
               log,
               prefix=None,
               aggregation_fields=None,
               time_field=None,
               target_field=None,
               target_steps=None,
               verbose=True,
               past_binarized_src_full_path=None,
               timers_section=None,
               **kwargs):
    self._fn_metadata = 'metadata.json'
    self._fn_mappings = 'mappings.pickle'

    super(TimeseriesDataMaster, self).__init__(
      log=log,
      prefix=prefix,
      verbose=verbose,
      past_binarized_src_full_path=past_binarized_src_full_path,
      **kwargs
    )
    self._timers_section = timers_section
    ### attributes that will be set in `set_data_column_names`
    self.aggregation_fields = None
    self.time_field = None
    self.target_field = None
    self.target_steps = None
    ###
    self.series2agg = None
    self.agg2series = None
    self.series2aggfields = None
    self.dct_metadata = None
    self.datasets = {}
    self._nr_create_calls = 0
    self._mappings_loaded = False

    self.set_data_column_names(
      aggregation_fields=aggregation_fields,
      time_field=time_field,
      target_field=target_field,
      target_steps=target_steps
    )

    self.dct_metadata = {
      NumpyizeConstants.KEY_DATASET : {},
      NumpyizeConstants.KEY_DATASET_ORIGIN : {},
      NumpyizeConstants.KEY_AGGREGATION_FIELDS : self.aggregation_fields,
      NumpyizeConstants.KEY_TIME_FIELD : self.time_field,
      NumpyizeConstants.KEY_TARGET_FIELD: self.target_field,
      NumpyizeConstants.KEY_OTHER_COVARIATE_FIELDS : [],
      NumpyizeConstants.KEY_FREQ: None,
      NumpyizeConstants.KEY_NR_SERIES: None,
      NumpyizeConstants.KEY_NR_STEPS: None,
      NumpyizeConstants.KEY_TARGET_STEPS: self.target_steps
    }

    if self.prefix is not None:
      self.load_metadata(self.prefix)

    assert self.aggregation_fields is not None
    assert type(self.aggregation_fields) is list
    assert self.time_field is not None
    assert self.target_field is not None

    if self.prefix is not None:
      self.load_mappings(self.prefix)

    # for _ in self.dct_metadata[NumpyizeConstants.KEY_DATASET]:
    #   self.create_dataset()

    return

  def get_dataset_type(self, meta_info_id):
    return self.dct_metadata[NumpyizeConstants.KEY_DATASET][meta_info_id][NumpyizeConstants.KEY_DATASET_TYPE]

  def get_other_covariate_fields(self):
    return self.dct_metadata[NumpyizeConstants.KEY_OTHER_COVARIATE_FIELDS]
  
  def set_data_column_names(self,
                            aggregation_fields=None,
                            time_field=None,
                            target_field=None,
                            target_steps=None):
    """
    Method used for setting the data columns names. They are read directly
    from config data; otherwise, if specified here, they are overwritten
    """
    
    self._parse_config_data(
      aggregation_fields=aggregation_fields,
      time_field=time_field,
      target_field=target_field,
      target_steps=target_steps
    )
    return
  #enddef
  
  @staticmethod
  def time_mapping(start, end, freq='D'):
    date_range = pd.date_range(
      start=start,
      end=end,
      freq=freq
    )
    
    dct_map = {k:i for i, k in enumerate(date_range)}
    return date_range, dct_map
  #enddef
  
  def compute_mappings(self, df_chunk):    
    self.log.start_timer('_compute_mappings', section=self._timers_section)
    
    if not self.series2agg:
      self.series2agg = []
    
    if not self.agg2series:
      self.agg2series = {}
      
    if not self.series2aggfields:
      self.series2aggfields = {}
      for f in self.aggregation_fields:
        self.series2aggfields[f] = []

    self.log.start_timer('groupby_aggregation_fields_1', section=self._timers_section)
    keys = df_chunk.groupby(self.aggregation_fields).indices.keys()
    self.log.end_timer_no_skip('groupby_aggregation_fields_1', section=self._timers_section)
    set_series = set(self.series2agg)

    self.series2agg.extend([x for x in keys if x not in set_series])
    self.agg2series = {v:i for i,v in enumerate(self.series2agg)}

    for i, f in enumerate(self.aggregation_fields):
      self.series2aggfields[f] = [tuple_scalar(x)[i] for x in self.series2agg]
    #endfor

    self.log.end_timer_no_skip('_compute_mappings', section=self._timers_section)
    self.P("Master computed mappings... describe:")
    self._describe()
    return
  #enddef
  
  def save_mappings(self, prefix=None):
    if not prefix:
      prefix = self.log.file_prefix

    all_mappings = {
      'series2agg' : self.series2agg,
      'agg2series' : self.agg2series,
      'series2aggfields' : self.series2aggfields,
    }
    
    fn = os.path.join(prefix, self._fn_mappings)
    self.log.save_pickle_to_output(all_mappings, fn)
    return
  #enddef
  
  def load_mappings(self, prefix):
    datafiles = os.listdir(self.get_dir(prefix))
    datafiles = [x for x in datafiles if x == self._fn_mappings]
    
    if len(datafiles) > 1:
      raise ValueError("Datafiles content: {}".format(datafiles))
    elif len(datafiles) == 0:
      return
    
    fn = os.path.join(prefix, datafiles[0])
    all_mappings = self.log.load_pickle_from_output(fn, verbose=self.verbose)
    
    self.series2agg       = all_mappings['series2agg']
    self.agg2series       = all_mappings['agg2series']
    self.series2aggfields = all_mappings['series2aggfields']
    
    self._mappings_loaded = True
    self.P("Master loaded mappings... describe:")
    self._describe()
    return
  #enddef

  def save_metadata(self, prefix=None):
    if not prefix:
      prefix = self.log.file_prefix

    fn = os.path.join(prefix, self._fn_metadata)
    self.log.save_output_json(self.dct_metadata, fn)
    return

  def load_metadata(self, prefix):
    datafiles = os.listdir(self.get_dir(prefix))
    datafiles = [x for x in datafiles if x == self._fn_metadata]

    if len(datafiles) > 1:
      raise ValueError("Datafiles content: {}".format(datafiles))
    elif len(datafiles) == 0:
      return

    fn = os.path.join(prefix, datafiles[0])
    self.dct_metadata = self.log.load_output_json(fn, verbose=self.verbose)

    self.aggregation_fields = self.dct_metadata[NumpyizeConstants.KEY_AGGREGATION_FIELDS]
    self.time_field = self.dct_metadata[NumpyizeConstants.KEY_TIME_FIELD]
    self.target_field = self.dct_metadata[NumpyizeConstants.KEY_TARGET_FIELD]
    self.target_steps = self.dct_metadata[NumpyizeConstants.KEY_TARGET_STEPS]
    return

  
  def get_nr_entities_per_each_agg(self):
    dct = {}
    for f in self.aggregation_fields:
      nr = len(set(self.series2aggfields[f]))
      dct[f] = nr
    #endfor
    return dct
  

  def _describe(self):
    if not self.verbose:
      return

    self.Pnp("Master sumar statistics:")
    self.Pnp("* Total number of series: {:,}".format(len(self.series2agg)))
    dct_nr_entities = self.get_nr_entities_per_each_agg()
    for f,nr in dct_nr_entities.items():
      self.Pnp("* Total number of '{}': {:,}".format(f, nr))
    self.Pnp("\n")
    return
  #enddef
  
  def create_dataset(self):
    ds = TimeseriesDataMaster.Dataset(
      log=self.log,
      outer=self,
      meta_info_id=self._nr_create_calls,
      DEBUG=self.DEBUG
    )
    
    self.datasets[self._nr_create_calls] = ds
    self._nr_create_calls += 1
    return ds
  #enddef
  
  class Dataset(DecentrAIObject):  
    def __init__(self, log, outer, meta_info_id=0, **kwargs):
      super().__init__(log=log, prefix_log='[DSDATAB]', **kwargs)
      self.outer = outer
      self.verbose = self.outer.verbose
      self.meta_info_id = meta_info_id
      self.is_master = self.meta_info_id == 0
      self.freq = None

      self.dataset_type = None
      self.filename = None
      
      self.global_min_dt = None
      self.global_max_dt = None
      self.series_min_dt = {}
      self.series_ids_min_dt = {}
      self.series_max_dt = {}
      self.series_ids_max_dt = {}
      self.date_range = []
      self.dates2id = {}
      self.strdates2id = {}
      
      self.connector = None
      self.filters = None
      self.ffill_bfill_columns = None
      
      self.column_names = None
      
      self._fn_preprocess = 'preprocess_{}.pickle'
      self._fn_series = 'series_{}.hdf5'
      
      self._load_preprocessed_data(prefix=self.outer.prefix)

      return
    #enddef
    
    def _not_preprocessed(self):
      not_preprocessed = self.global_min_dt is None
      not_preprocessed = self.global_max_dt is None or not_preprocessed
      not_preprocessed = len(self.series_min_dt) == 0 or not_preprocessed
      not_preprocessed = len(self.series_max_dt) == 0 or not_preprocessed
      not_preprocessed = len(self.date_range) == 0 or not_preprocessed
      not_preprocessed = len(self.dates2id) == 0 or not_preprocessed
      return not_preprocessed
    #enddef
    
    def star_series(self, np_datasets, count=0, lint=None, rint=None,
                    np_mappings=None, min_dt=None):
      if lint is not None:
        assert type(lint) is str
        lint = self.strdates2id[lint]
      
      if rint is not None:
        assert type(rint) is str
        rint = self.strdates2id[rint]
        rint += 1
        
      arr = None
      if type(np_datasets) is dict:
        for k,v in np_datasets.items():
          break
        arr = v
      elif type(np_datasets) in np.ndarray:
        arr = np_datasets
      else:
        raise ValueError('Unknown np_datasets')
    
      if np_mappings is None:
        np_mappings = np.arange(arr.shape[0])
    
      good_starts = [True for _ in range(np_mappings.shape[0])]
      
      if min_dt is not None:
        starts = [self.series_ids_min_dt[i] for i in np_mappings]
        good_starts = list(map(lambda x: x <= pd.to_datetime(min_dt), starts))
      #endif
      
      good_starts = np.array(good_starts)
      mask_good_series = (arr[:, lint:rint] > 0).sum(axis=1) >= count
      mask_good_series = mask_good_series & good_starts
      ### TODO debug for _lens_data/_v2/_adph_dev_other_covariates - weekly-tran-v3
      
      mask_star_series = ~mask_good_series
      idx = np.where(mask_star_series)[0]
      
      return np_mappings[idx], mask_star_series
    
    
    def train_test_split(self, np_datasets, lint_train=None,
                         rint_train=None, lint_test=None,
                         rint_test=None):
      
      
      def _split(arr):
        if len(arr.shape) == 1:
          arr = np.expand_dims(arr, -1)
        _arr_train = arr[:,lint_train:rint_train]
        _arr_test = None
        if lint_test or rint_test:
          _arr_test = arr[:, lint_test:rint_test]
        return _arr_train, _arr_test
      
      np_train_datasets, np_test_datasets = None, None

      if lint_train is not None:
        assert type(lint_train) in [str, int, np.int32, np.int16, np.int64]
        if type(lint_train) is str:
          lint_train = self.strdates2id[lint_train]
      
      if rint_train is not None:
        assert type(rint_train) in [str, int, np.int32, np.int16, np.int64]
        if type(rint_train) is str:
          rint_train = self.strdates2id[rint_train]
          rint_train += 1
        
      if lint_test is not None:
        assert type(lint_test) in [str, int, np.int32, np.int16, np.int64]
        if type(lint_test) is str:
          lint_test = self.strdates2id[lint_test]
        
      if rint_test is not None:
        assert type(rint_test) in [str, int, np.int32, np.int16, np.int64]
        if type(rint_test):
          rint_test = self.strdates2id[rint_test]
          rint_test += 1
        
        
      if type(np_datasets) is dict:
        np_train_datasets = {}
        
        if lint_test or rint_test:
          np_test_datasets = {}
        
        for k,v in np_datasets.items():
          arr_train, arr_test = _split(v)
          np_train_datasets[k] = arr_train
          if type(np_test_datasets) is dict:
            np_test_datasets[k] = arr_test
          #endif
        #endfor
      elif type(np_datasets) is np.ndarray:
        np_train_datasets, np_test_datasets = _split(np_datasets)
      else:
        raise ValueError('Unknown np_datasets')
      
      return np_train_datasets, np_test_datasets

    
    def __call__(self,
                 lint_train=None, rint_train=None,
                 lint_test=None, rint_test=None,
                 lint_star_not_predictible=None,
                 rint_star_not_predictible=None,
                 count_star_not_predictible=0,
                 min_dt_predictible=None,
                 sanity=False):
      
      if self._not_preprocessed():
        self._load_preprocessed_data(prefix=self.outer.prefix)

      datasets = self._load_sparse_series(
        prefix=self.outer.prefix,
        sanity=sanity
      )
      
      lst_np_datasets,\
        lst_np_mappings, _ = self._sparse_series_to_array(*datasets)
        
      lst_final_np_datasets = []
      for np_datasets in lst_np_datasets:
        final_np_datasets = {}
        for col,v in np_datasets.items():
          if col in self.ffill_bfill_columns:
            # self.P("Applying ffill / bfill on '{}' array ...".format(col))
            final_np_datasets[col] = self.log.np_bfill(self.log.np_ffill(v))
            # self.P("", t=True)
          else:
            final_np_datasets[col] = v
          #endif
        #endfor
        lst_final_np_datasets.append(final_np_datasets)
      #endfor
      
      # stars = []
      lst_train_np_datasets = []
      lst_test_np_datasets = []
      for i in range(len(lst_final_np_datasets)):
        # stars.append(self.star_series(
        #   np_datasets=lst_final_np_datasets[i],
        #   count=count_star_not_predictible,
        #   lint=lint_star_not_predictible,
        #   rint=rint_star_not_predictible,
        #   np_mappings=lst_np_mappings[i],
        #   min_dt=min_dt_predictible
        # ))
        
        np_train_datasets, np_test_datasets = self.train_test_split(
          np_datasets=lst_final_np_datasets[i],
          lint_train=lint_train,
          rint_train=rint_train,
          lint_test=lint_test,
          rint_test=rint_test
        )
        
        lst_train_np_datasets.append(np_train_datasets)
        lst_test_np_datasets.append(np_test_datasets)
      #endfor
      
      ret_train_datasets = lst_train_np_datasets
      ret_test_datasets  = lst_test_np_datasets
      ret_mappings = lst_np_mappings
      # ret_stars = stars
      # ret_tmp = lst_final_np_datasets
      
      if len(lst_train_np_datasets) == 1:
        ret_train_datasets = lst_train_np_datasets[0]
        ret_test_datasets = lst_test_np_datasets[0]
        ret_mappings = lst_np_mappings[0]
        # ret_stars = stars[0]
        # ret_tmp = lst_final_np_datasets[0]
      #endif
  
      return (
              ret_train_datasets,
              ret_test_datasets,
              ret_mappings,
              # ret_stars,
              # ret_tmp,
             )
    #enddef
    
    def setup(self,
              dataset_type=None,
              column_names=None,
              filters=None,
              ffill_bfill_columns=None,
              freq=None,
              series_chunksize=None,
              connector_type=ConnectorConstants.DF_CONNECTOR,
              column_sanity_check=None,
              force=False,
              **conn_kwargs):
      ### freq='D' / 'W-MON'
      if dataset_type is None:
        dataset_type = NumpyizeConstants.DATASET_TRAIN

      if freq is None:
        freq = self.outer.dct_metadata[NumpyizeConstants.KEY_FREQ]

      assert freq is not None
      assert isinstance(freq, str)
      assert dataset_type in NumpyizeConstants.VALID_DATASET_NAMES

      if ffill_bfill_columns is None:
        ffill_bfill_columns = []
      if filters is None:
        filters = []
      if column_names is None:
        column_names = []
      
      self.freq = freq.upper()
      self.filters = filters
      self.column_names = column_names
      self.ffill_bfill_columns = ffill_bfill_columns
      self.dataset_type = dataset_type

      assert type(self.column_names) is list
      assert type(self.ffill_bfill_columns) is list
      self.ffill_bfill_columns = list(set(self.ffill_bfill_columns) & set(self.column_names))

      ### if we add multiple EXTRA_PROCESSING params such as ffill_bfill_columns, then do the intersect with self.column_names
      
      self.keep_columns = self.outer.aggregation_fields +\
                          [self.outer.time_field] +\
                          self.column_names
      
      if not self.outer._mappings_loaded or force:
        self.outer.dct_metadata[NumpyizeConstants.KEY_DATASET][self.meta_info_id] = {}

        self.connector = self._create_connector(
          connector_type=connector_type,
          **conn_kwargs
        )

        self.outer.dct_metadata[NumpyizeConstants.KEY_DATASET][self.meta_info_id][NumpyizeConstants.KEY_DATASET_TYPE] = self.dataset_type
        self.preprocess()
        self.process(
          series_chunksize=series_chunksize,
          column_sanity_check=column_sanity_check
        )

        self.outer.save_metadata(prefix=self.outer.prefix)



      #endif

      return
    #enddef

    def __repr__(self):
      str_descr  = '* dataset type: {}'.format(self.dataset_type)
      str_descr += '\n* targets: {}'.format(self.column_names)
      str_descr += "\n* frequency='{}'".format(self.freq)
      str_descr += '\n* history between {} and {}'.format(
        self.global_min_dt, self.global_max_dt
      )
      
      return str_descr
    
    
    def _create_connector(self,
                          connector_type=ConnectorConstants.CSV_CONNECTOR,
                          **kwargs):
      """
      Setups the data connector to the dataset
      
      Parameters:
      ----------
      connector_type : str, optional
        Specifies how the data should be consumed:
          - const.CSV_CONNECTOR ('CSV'): consumes data from a csv file
          - const.ODBC_CONNECTOR ('ODBC'): consumes data from an ODBC connector
        The default is 'CSV'.
  
      **kwargs
        Contain configurations for the data connector.
        The **kwargs can have the following parameters:
          1. `dataset` : str, optional
            The dataset name as described in the data config file.
            Used for [CSV / ODBC].
            The default is None
          
          2. `fn_data` : str, optional
            The path to the dataset (if specified, overwrites the one from the
                                     config file).
            Used for [CSV].
            The default is None
          
          3. `chunksize` : int optional
            Number of data row processed per chunk
            Used for [CSV / ODBC].
            The default is None
            
          4. `parse_dates` : bool, optional
            Sets if the dataset time field should be casted to DateTime
            Used for [CSV / ODBC]
            The default is True
        
          5. `database` : str, optional
            specifies what databese schema should be used (if specified,
            overwrites the one from the config file).
            Used for [ODBC]
            The default is None.
        
          6. `uid` : str, optional 
            the username used for the database connection (if specified,
            overwrites the one from the config file).
            Used for [ODBC]
            The default is None.
          
          7. `password` : str, optional 
            the password used for the database connection (if specified,
            overwrites the one from the config file).
            Used for [ODBC]
            The default is None.
            
          8.  `odbc_driver` : str, optional 
            the driver to use for connecting to the database (if specified,
            overwrites the one from the config file).
            Used for [ODBC].
            The default is None.
            
          9. `port` : str, optional
            port used in the database connection (if specified,
            overwrites the one from the config file).
            Used for [ODBC].
            The default is ''.
            
          10. `table_data` : str, optional 
            from which table to extract the data (if specified,
            overwrites the one from the config file).
            Used for [ODBC]
            The default is None.
            
          11. `server` : str, optional
            the server which hosts the database (if specified,
            overwrites the one from the config file).
            Used for [OBDC]
            The default is None.
            
          12. `DEBUG` : bool, optional
            DEBUG flag in connectors that enables additional logs
            Used for [CSV / ODBC]
            The default is True.

          13. `absolute_path_fn_data` : bool, optional
            Flag that specifies if `fn_data` is an absolute path or not. If not, it looks
            for `fn_data` in logger's _data main subfolder
            Used for [CSV]
            The default is False
        
      Returns
      -------
      a connector object
      """

      DEBUG = kwargs.get('DEBUG', True)
      assert connector_type.upper() in [
        ConnectorConstants.CSV_CONNECTOR,
        ConnectorConstants.ODBC_CONNECTOR,
        ConnectorConstants.DF_CONNECTOR
      ]
      
      if connector_type.upper() == ConnectorConstants.CSV_CONNECTOR:
        dataset = kwargs.get('dataset', None)
        chunksize = kwargs.get('chunksize', None)
        bool_parse_dates = kwargs.get('parse_dates', True)

        fn_data = kwargs.get('fn_data', None)
        absolute_path_fn_data = kwargs.get('absolute_path_fn_data', False)

        if fn_data is not None:
          self.filename = fn_data.split('/')[-1]

        if self.meta_info_id not in self.outer.dct_metadata[NumpyizeConstants.KEY_DATASET]:
          self.outer.dct_metadata[NumpyizeConstants.KEY_DATASET][self.meta_info_id] = {}

        if self.meta_info_id not in self.outer.dct_metadata[NumpyizeConstants.KEY_DATASET_ORIGIN]:
          self.outer.dct_metadata[NumpyizeConstants.KEY_DATASET_ORIGIN][self.meta_info_id] = {}

        self.outer.dct_metadata[NumpyizeConstants.KEY_DATASET_ORIGIN][self.meta_info_id][
          NumpyizeConstants.KEY_DATASET_FN_DATA] = fn_data
        self.outer.dct_metadata[NumpyizeConstants.KEY_DATASET_ORIGIN][self.meta_info_id][
          NumpyizeConstants.KEY_DATASET_ABSOLUTE_PATH_FN_DATA] = absolute_path_fn_data
        self.outer.dct_metadata[NumpyizeConstants.KEY_DATASET_ORIGIN][self.meta_info_id][
          NumpyizeConstants.KEY_DATASET_FILENAME] = self.filename

        parse_dates = None
        if bool_parse_dates and self.outer.time_field:
          parse_dates = [self.outer.time_field]

        c = connectors.CSVConnector(log=self.log, DEBUG=DEBUG)
        c.set_data_connector_params(
          dataset=dataset,
          fn_data=fn_data,
          chunksize=chunksize,
          parse_dates=parse_dates,
          absolute_path_fn_data=absolute_path_fn_data
        )
  
      elif connector_type.upper() == ConnectorConstants.ODBC_CONNECTOR:
        database = kwargs.get('database', None)
        uid = kwargs.get('uid', None)
        password = kwargs.get('password', None)
        odbc_driver = kwargs.get('odbc_driver', None)
        port = kwargs.get('port', '')
        table_data = kwargs.get('table_data', None)
        server = kwargs.get('server', None)
        chunksize = kwargs.get('chunksize', None)
        c = connectors.ODBCConnector(log=self.log, DEBUG=DEBUG)
        c.set_data_connector_params(
          server=server,
          database=database, uid=uid, password=password,
          odbc_driver=odbc_driver, port=port,
          table_data=table_data, chunksize=chunksize
        )

      elif connector_type.upper() == ConnectorConstants.DF_CONNECTOR:
        df = kwargs['df']
        c = connectors.DFConnector(log=self.log, DEBUG=DEBUG)
        c.set_data_connector_params(df=df)
      #endif

      return c

    def _compute_min_max_dt(self, df_chunk):
      self.log.start_timer('_compute_min_max_dt', section=self.outer._timers_section)
      crt_min_time = df_chunk[self.outer.time_field].min()
      crt_max_time = df_chunk[self.outer.time_field].max()
      
      self.global_min_dt = min(self.global_min_dt, crt_min_time) \
                           if self.global_min_dt is not None \
                           else crt_min_time
                           
      self.global_max_dt = max(self.global_max_dt, crt_max_time) \
                           if self.global_max_dt is not None \
                           else crt_max_time
      self.log.end_timer_no_skip('_compute_min_max_dt', section=self.outer._timers_section)
      return
    #enddef

    def _auto_compute_weekly_freq(self):
      day_of_week = self.global_min_dt.day_of_week

      dct_day_of_week = {
        0 : 'W-MON', 1 : 'W-TUE', 2 : 'W-WED', 3 : 'W-THU',
        4 : 'W-FRI', 5 : 'W-SAT', 6 : 'W-SUN'
      }

      return dct_day_of_week[day_of_week]


    
    def _describe(self):
      """
      Prints basic information about the current loaded preprocess data:
  
      Constrains:
      -----------
        Must be called after preprocess
      """
      if not self.verbose:
        return

      P = self.outer.Pnp
      
      P("Dataset {} sumar statistics:".format(self.meta_info_id))
      P(repr(self))

      random_series_indexes = np.random.choice(
        range(len(self.outer.series2agg)),
        replace=False,
        size=5
      )

      P("Printing random series ...")
      P("  - min_dt   = the first day of transaction")
      P("  - max_dt   = the last day of transaction")
      for idx in np.sort(random_series_indexes):
        fields = self.outer.series2agg[idx]
        if fields not in self.series_min_dt:
          continue

        min_dt = str(self.series_min_dt[fields])[:10]
        max_dt = str(self.series_max_dt[fields])[:10]
        str_fields = '[{}]'.format(
          '_'.join(map(lambda x:str(x), tuple_scalar(fields)))
        )
  
        P("Series #{:>9} (aggregation {:>15}) - min_dt=[{}] / max_dt=[{}]".format(
          idx, str_fields, min_dt, max_dt
        ))

      P("\n")
      return
    #enddef
    
    def _data_check(self, df_chunk):
      """
      Checks the data for `inf` or `nan` values and raises a `ValueError`
      if any found
  
      Parameters:
      ----------
        `df_chunk`: the data chunk to be checked
  
      Returns:
      -------
        None
      """
      
      def _has_nulls(col):
        return df_chunk[col].isnull().values.any()
      
      def _has_infs(col):
        return np.inf in df_chunk[col].unique()
      
      for column in df_chunk.columns:
        if np.issubdtype(df_chunk[column].dtype, np.number):
          if _has_infs(column):
            raise ValueError
        
        if column not in self.ffill_bfill_columns:
          if _has_nulls(column):
            self.ffill_bfill_columns.append(column)
            self.P("WARNING! Column '{}' added automatically to ffill_bfill_columns because it contains nans!".format(column), color='r')
        
      return
    #enddef
    
    def _filter_series(self, df_chunk, *filters):
      """
      Applies basic filters over a chunk of data to discard unwanted data
  
      Parameters:
      -----------
        1. `df_chunk`: the data chunk to be filtered
        ......
        (column, string operator, value, [optional convert to datetime])
  
      Returns:
      --------
        Filtered chunk
  
      """
      
      dct_operators = {
        "<"    : operator.lt,
        "<="   : operator.le,
        "=="   : operator.eq,
        "!="   : operator.ne,
        ">="   : operator.ge,
        ">"    : operator.gt,
        "isin" : lambda x,v: x.isin(v)
      }

      if len(filters) == 0:
        return df_chunk

      self.log.start_timer('_filter_series', section=self.outer._timers_section)
      self.P("Filtering series chunk with [{}] ...".format(filters))

      for arg in filters:
        assert len(arg) >= 3
        convert_to_datetime = False
        column = arg[0]
        string_op = arg[1]
        assert string_op in dct_operators
        value = arg[2]
        if len(arg) > 3:
          convert_to_datetime = arg[3]
        
        if convert_to_datetime:
          value = pd.to_datetime(value)
        
        op = dct_operators[string_op]
        
        df_chunk = df_chunk[op(df_chunk[column], value)]
      #endfor
      self.P("", t=True)
      self.log.end_timer_no_skip('_filter_series', section=self.outer._timers_section)
      return df_chunk
    #enddef
    
    def _save_preprocessed_data(self, prefix=None):
      """
      Saves the preprocessed data to disk.

      Parameters:
      -----------
        1. `prefix`: Prefix of the filename.
        If it's not specified `Logger` will pick one for it

      Returns:
      --------
        None

      Constrains:
      -----------
        Must be called after preprocess
      """
      if not prefix:
        prefix = self.log.file_prefix
  
      preprocessed_data = {
        'filters' : self.filters,
        'freq'    : self.freq,
        'column_names' : self.column_names,
        'ffill_bfill_columns' : self.ffill_bfill_columns,
        'dataset_type' : self.dataset_type,
        'filename' : self.filename,

        'global_min_dt' : self.global_min_dt,
        'global_max_dt' : self.global_max_dt,
        'series_min_dt' : self.series_min_dt,
        'series_max_dt' : self.series_max_dt,
      }
      
      fn = os.path.join(
        prefix,
        self._fn_preprocess.format(self.meta_info_id)
      )
      self.log.save_pickle_to_output(preprocessed_data, fn)

      return
    #enddef
    
    def _load_preprocessed_data(self, prefix):
      """
        Loads the preprocessed data from disk.
  
        Parameters:
          1. `prefix`: Prefix of the filename.
  
        Returns:
          None
      """
      datafiles = os.listdir(self.outer.get_dir(prefix))
      datafiles = [x for x in datafiles if
                   x == self._fn_preprocess.format(self.meta_info_id)]
  
      if len(datafiles) > 1:
        raise ValueError("Datafiles content: {}".format(datafiles))
      elif len(datafiles) == 0:
        return False
      
      fn = os.path.join(prefix, datafiles[0])
      preprocessed_data = self.log.load_pickle_from_output(fn, verbose=self.verbose)
      
      self.filters       = preprocessed_data['filters']
      self.freq          = preprocessed_data['freq']
      self.column_names  = preprocessed_data['column_names']
      self.global_min_dt = preprocessed_data['global_min_dt']
      self.global_max_dt = preprocessed_data['global_max_dt']
      self.series_min_dt = preprocessed_data['series_min_dt']
      self.series_max_dt = preprocessed_data['series_max_dt']
      self.ffill_bfill_columns = preprocessed_data['ffill_bfill_columns']
      self.dataset_type = preprocessed_data.get('dataset_type', None)
      self.filename = preprocessed_data.get('filename', None)
      
      self._describe()
      self._create_date_range()
      self._compute_series_ids_min_max_dt()
      return True
    #enddef
    
    def _create_date_range(self):
      self.date_range, self.dates2id = self.outer.time_mapping(
        start=self.global_min_dt,
        end=self.global_max_dt,
        freq=self.freq
      )
      self.strdates2id = {
        k.strftime('%Y-%m-%d'):v for k,v in self.dates2id.items()
      }
      return
    
    def preprocess(self):
      """
      The preprocess method computes all the preprocessing for the data
      and needs configuration for the data connector.
      In the end saves the preprocess data to the disk.
  
      """

      self.P("Start - Preprocess step for dataset {}".format(self.meta_info_id), color='b')

      gen = self.connector.chunk_generator()
      for itr, df_chunk in enumerate(gen):
        self.P("Start - Chunk {} [{}] ...".format(itr, df_chunk.shape))

        if itr == 0:
          if is_void_list(self.column_names):
            all_columns = df_chunk.columns.to_list()
            self.column_names = list(set(all_columns) - set(self.outer.aggregation_fields + [self.outer.time_field]))
            self.keep_columns += self.column_names
          #endif

          other_covariate_columns = list(set(self.column_names) - set([self.outer.target_field]))
          self.outer.dct_metadata[NumpyizeConstants.KEY_OTHER_COVARIATE_FIELDS] += other_covariate_columns
          self.outer.dct_metadata[NumpyizeConstants.KEY_OTHER_COVARIATE_FIELDS] = list(set(self.outer.dct_metadata[NumpyizeConstants.KEY_OTHER_COVARIATE_FIELDS]))
        #endif
        
        df_chunk = df_chunk[self.keep_columns]
        self._data_check(df_chunk)
        df_chunk = self._filter_series(df_chunk, *self.filters)
        
        self._compute_min_max_dt(df_chunk)

        if self.freq == 'W':
          self.freq = self._auto_compute_weekly_freq()

        if self.is_master:
          self.outer.compute_mappings(df_chunk)

        self._compute_series_min_max_dt(df_chunk)
        
        self.P("End - Chunk {} [{}].".format(itr, df_chunk.shape))
      #endfor
      
      self._compute_series_ids_min_max_dt()
      self._create_date_range()

      if self.is_master:
        self.outer.dct_metadata[NumpyizeConstants.KEY_FREQ] = self.freq
        self.outer.dct_metadata[NumpyizeConstants.KEY_NR_SERIES] = len(self.outer.series2agg)
        self.outer.dct_metadata[NumpyizeConstants.KEY_NR_STEPS] = self.date_range.shape[0]
        self.outer.save_mappings(prefix=self.outer.prefix)

      self._save_preprocessed_data(prefix=self.outer.prefix)

      self.P("Dataset {} preprocessed ... describe:".format(self.meta_info_id))
      self._describe()

      self.P("#" * 30)
      self.log.show_timers()
      self.P("#" * 30)

      self.P("End - Preprocess step for dataset {}".format(self.meta_info_id), color='b')
      return
    #enddef
    
    def _compute_series_ids_min_max_dt(self):
      self.series_ids_min_dt = {
        self.outer.agg2series.get(k, None) : v for k,v in self.series_min_dt.items()
      }
      self.series_ids_max_dt = {
        self.outer.agg2series.get(k, None) : v for k,v in self.series_max_dt.items()
      }
      self.series_ids_min_dt.pop(None, None)
      self.series_ids_max_dt.pop(None, None)
      return
    
    def _compute_series_min_max_dt(self, df_chunk):
      self.P("Computing min/max datetime per each serie ...")
      self.log.start_timer('_compute_series_min_max_dt', section=self.outer._timers_section)
      self.log.start_timer('groupby_aggregation_fields_2', section=self.outer._timers_section)
      grp = df_chunk.groupby(self.outer.aggregation_fields)
      self.log.end_timer_no_skip('groupby_aggregation_fields_2', section=self.outer._timers_section)
  
      # compute min dt
      dct_min_dt = grp[self.outer.time_field].min().to_dict()
      self.series_min_dt = {
        x : min(self.series_min_dt.get(x, pd.Timestamp.max),
                dct_min_dt.get(x, pd.Timestamp.max))
        for x in set(self.series_min_dt.keys()) | set(dct_min_dt.keys())
      }
  
      # compute max dt
      dct_max_dt = grp[self.outer.time_field].max().to_dict()
      self.series_max_dt = {
        x : max(self.series_max_dt.get(x, pd.Timestamp.min),
                dct_max_dt.get(x, pd.Timestamp.min))
        for x in set(self.series_max_dt.keys()) | set(dct_max_dt.keys())
      }
      
      self.log.end_timer_no_skip('_compute_series_min_max_dt', section=self.outer._timers_section)
      return
    #enddef
    
    def process(self, series_chunksize=None, column_sanity_check=None):
    
      nr_series = len(self.outer.series2agg)
      if not series_chunksize:
        series_chunksize = nr_series

      self.P("Start - Process step for dataset {}".format(self.meta_info_id), color='b')
  
      for series_chunk_idx in range(0, nr_series, series_chunksize):
        gen = self.connector.chunk_generator()
        data = None
  
        self.log.start_timer('process_dataset_pass', section=self.outer._timers_section)
        for itr, df_chunk in enumerate(gen):
          self.P("Start - Chunk #{} [{}] ...".format(itr, df_chunk.shape))
          df_chunk = df_chunk[self.keep_columns]
          df_chunk = self._filter_series(df_chunk, *self.filters)
          df_chunk = self._map_aggregation2series(df_chunk)
          
          df_chunk = self._select_subseries(
            df_chunk,
            series_chunk_idx=series_chunk_idx,
            series_chunksize=series_chunksize
          )
          
          # in case none of the subseries are in the current chunk, continue
          if df_chunk.empty:
            continue
          
          if data is None:
            data = df_chunk
          else:
            data = data.append(df_chunk)
          #endif
          self.P("End - chunk #{} [{}].".format(itr, df_chunk.shape))
        #endfor
        self.log.end_timer_no_skip('process_dataset_pass', section=self.outer._timers_section)
        
        if self.is_master: #TODO
          data = self._normalize_series(data)
        
        dataset = self._parse_series_sparse(
          data,
          *self.column_names,
          column_sanity_check=column_sanity_check
        )
  
        index = 'subset_' + str(series_chunk_idx//series_chunksize)
        write_mode = 'w'
        if series_chunk_idx > 0:
          write_mode = 'a'

        self._save_sparse_series(
          index=index,
          dataset=dataset,
          write_mode=write_mode,
          prefix=self.outer.prefix
        )
        self.P("#" * 30)
        self.log.show_timers()
        self.P("#" * 30)

        self.P("End - Process step for dataset {}".format(self.meta_info_id), color='b')
      return
    #enddef
    
    def _map_aggregation2series(self, df_chunk):
      """
      Maps aggregation to series_id for each data point
  
      Parameters:
      -----------
        1. `df_chunk`: DataFrame to be mapped.
  
      Returns:
      --------
        DataFrame
  
      Constraints:
      ------------
        The `chunk` parameter must have `aggregation_fields` columns
  
      """
      self.log.start_timer('_map_aggregation2series', section=self.outer._timers_section)
      self.P("  Mapping aggregation to series_ids ...")
      df_chunk = df_chunk.set_index(self.outer.aggregation_fields)
      series_index = df_chunk.index.to_series().map(self.outer.agg2series)
      df_chunk[NumpyizeConstants.COLUMN_SERIES_INDEX] = series_index
      if df_chunk[NumpyizeConstants.COLUMN_SERIES_INDEX].hasnans:
        df_chunk.dropna(subset=[NumpyizeConstants.COLUMN_SERIES_INDEX], inplace=True)
        df_chunk[NumpyizeConstants.COLUMN_SERIES_INDEX] = df_chunk[NumpyizeConstants.COLUMN_SERIES_INDEX].astype(np.int32)

      self.P("  ", t=True)
      self.log.end_timer_no_skip('_map_aggregation2series', section=self.outer._timers_section)
      return df_chunk.reset_index()
    #enddef
    
    def _select_subseries(self, df_chunk, series_chunk_idx, series_chunksize):
      """
      From the given data, selects only the data points with series_id
        between `subseries_index` and `subseries_index` +
        `subseries_chunk_size`
  
      Parameters:
      -----------
        1. `chunk`: DataFrame to be mapped.
        2. `subseries_index`: The first series_id of the current
        subseries chunk
        3. `subseries_chunk_size`: The size of each subseries chunk
  
      Returns:
      --------
        DataFrame
  
      Constraints:
      ------------
        `df_chunk` parameter must have column `const.COLUMN_SERIES_INDEX`
      """
      
      self.log.start_timer('select_subseries', section=self.outer._timers_section)
      self.P("  Selecting subseries from index {} to {} ..."
         .format(series_chunk_idx, series_chunk_idx+series_chunksize))
      range_finish = min(
        series_chunk_idx + series_chunksize,
        len(self.outer.series2agg)+1
      )
      select = list(range(series_chunk_idx, range_finish))
      df_chunk = df_chunk[df_chunk[NumpyizeConstants.COLUMN_SERIES_INDEX].isin(select)]
      self.P("  ", t=True)
      self.log.end_timer_no_skip('select_subseries', section=self.outer._timers_section)
      return df_chunk.reset_index()
    #enddef
    
    def _normalize_series(self, df_chunk, *rules):
      """
      Aggregates multiple transactions per period:
        - Sums quantity
        - TODO in the future - other columns will have particular aggregation
        
      Parameters:
      -----------
        1. `df_chunk`: DataFrame to be normalized
  
      Returns:
      --------
        DataFrame
  
      Constraints:
      ------------
        `df_chunk` parameter must have columns `const.COLUMN_SERIES_INDEX`,
        `time_field`, `quantity_field`
      """
      # because we can have multiple rows for each series index and timestamp,
      # we need to aggregate the data
      self.log.start_timer('_normalize_series', section=self.outer._timers_section)
      self.P("Normalizing series data (multiple sales per step) ...")
      grp = df_chunk.groupby([NumpyizeConstants.COLUMN_SERIES_INDEX,
                              self.outer.time_field])
      df_grp = grp[self.column_names].sum() #TODO
      self.P("", t=True)
      self.log.end_timer('_normalize_series', section=self.outer._timers_section)
      return df_grp.reset_index()
    #enddef
    
    def _parse_series_sparse(self, df, *value_columns, column_sanity_check=None):
      """
       Splits a DataFrame into multiple data sets in order to be
       saved as sparce matrices
  
       Parameters:
         1. `partial_data`: DataFrame to be parsed
  
       Returns:
         `datasets` - a dictionary of data sets with the following elements:
            - quantity_field: a series of all transactions
            - price_field: a series of all prices
            - 'col_ind': the column index for each transaction
            - 'row_ind': the row index for each transaction
            - 'mapping': a map from row_ind to series_id
  
  
       Constraints:
         - `partial_data` must have columns `const.COLUMN_SERIES_INDEX`,
         `time_field`, `quantity_field`
         - needs `global_statistics` dict to be loaded in
         order to extract min_dt and max_dt
  
       """
      self.log.start_timer('_parse_series_sparse', section=self.outer._timers_section)
      self.P('Processing series as sparse matrices ...')
      
      df[NumpyizeConstants.COLUMN_TIME_ID] = df[self.outer.time_field].map(self.dates2id)
      self.log.start_timer('sort_values', section=self.outer._timers_section)
      df.sort_values(by=[NumpyizeConstants.COLUMN_SERIES_INDEX, self.outer.time_field],
                     inplace=True)
      self.log.end_timer_no_skip('sort_values', section=self.outer._timers_section)

      col_ind = df[NumpyizeConstants.COLUMN_TIME_ID]\
                  .values\
                  .astype(np.int32)
      mapping = df[NumpyizeConstants.COLUMN_SERIES_INDEX]\
                  .unique()\
                  .astype(np.int32)
      row_ind = df[NumpyizeConstants.COLUMN_SERIES_INDEX]\
                  .map({j:i for i,j in enumerate(mapping)})\
                  .values\
                  .astype(np.int32)

      dataset = {}
      for v in value_columns:
        assert v in df.columns
        values = df[v].values
        mask_non_zero = values != 0
        dataset[v] = {}

        dataset[v][NumpyizeConstants.SPARSE_VALUES]  = values[mask_non_zero]
        dataset[v][NumpyizeConstants.SPARSE_COL_IND] = col_ind[mask_non_zero]
        dataset[v][NumpyizeConstants.SPARSE_ROW_IND] = row_ind[mask_non_zero]
      #endfor

      dataset[NumpyizeConstants.SPARSE_MAPPING] = mapping
      dataset[NumpyizeConstants.SPARSE_FREQ] = self.freq
      
      self.P("", t=True)
      self.log.end_timer_no_skip('_parse_series_sparse', section=self.outer._timers_section)

      if column_sanity_check is None:
        if self.outer.target_field in dataset.keys():
          column_sanity_check = self.outer.target_field

      self._sanity_check_sparse_dataset(
        dataset,
        target=column_sanity_check,
        nr_series=3
      )

      return dataset
    #enddef
    
    def _sanity_check_sparse_dataset(self, dataset,
                                     target=None, nr_series=3):
      self.P("Sanity check sparse dataset:")
      self.P("-" * 30)
      
      if not target:
        target = self.column_names[0]
  
      P = partial(self.log.P, noprefix=True)
      sparse_row_ind = dataset[target][NumpyizeConstants.SPARSE_ROW_IND]
      sparse_col_ind = dataset[target][NumpyizeConstants.SPARSE_COL_IND]
      freq = dataset[NumpyizeConstants.SPARSE_FREQ]
      sparse_mapping = dataset[NumpyizeConstants.SPARSE_MAPPING]

      nr_lines = np.max(sparse_row_ind) + 1

      series_sparse_idx = np.random.choice(
        np.arange(nr_lines),
        replace=False,
        size=nr_series
      )
      
      series_real_id = sparse_mapping[series_sparse_idx]
      
      for i in range(nr_series):
        s_id = series_real_id[i]
        s_idx = series_sparse_idx[i]
        non_zero = np.where(sparse_row_ind == s_idx)[0]
        targets = dataset[target][NumpyizeConstants.SPARSE_VALUES][non_zero]
        time_ids = sparse_col_ind[non_zero]
        time_range, dct_map = self.outer.time_mapping(
          start=self.global_min_dt,
          end=self.global_max_dt,
          freq=freq
        )
        dct_inv_map = {v:k for k,v in dct_map.items()}
        timestamps = [dct_inv_map[t] for t in time_ids]
        
        nr_targets = targets.shape[0]
        lst_print = []
        for j in range(nr_targets):
          lst_print.append(
            timestamps[j].strftime("%Y-%m-%d") + " / {:>5.2f}".format(targets[j])
          )
        
        agg = self.outer.series2agg[s_id]
        P("Series id #{} ({}) [{} - {}]:".format(
          s_id,
          agg, 
          self.series_min_dt[agg].strftime('%Y-%m-%d'),
          self.series_max_dt[agg].strftime('%Y-%m-%d'))
        )
        
        self.log.print_on_columns(
          *lst_print,
          nr_print_columns=3,
          nr_print_chars=20
        )
        P("")
      #endfor
      return
    #enddef
    
    def _save_sparse_series(self,
                            index,
                            dataset,
                            write_mode='w',
                            prefix=None):
      """
      Saves a subset of the data as a hdf5 file.
  
      Parameters:
      -----------
        1. `index`: the name of the subset
        2. `dataset`: the data sets to be saved
        3. `prefix`: the file prefix, if None specified,
        the method will use log.file_prefix
  
      Returns:
      --------
        None
      """
      if not prefix:
        prefix = self.log.file_prefix
  
      fn = os.path.join(
        self.outer.get_dir(prefix),
        self._fn_series.format(self.meta_info_id)
      )

      self.P("Saving index '{}' of dataset to '{}' ...".format(index, fn))
      save_dict_to_hdf5(dataset, fn, root=index, write_mode=write_mode)
      return
    #enddef
    
    def _load_sparse_series_keys(self, prefix):
      """
      Loads the subset keys from a series file and returns them as a list.
      The file must have the name `prefix`_series_file.hdf5.
      If no prefix is specified, it will use log.file_prefix
  
      Parameters:
      -----------
        1. `prefix`: the filename prefix
  
      Returns:
      --------
        A list of subset keys
      """
      import h5py

      datafiles = os.listdir(self.outer.get_dir(prefix))
      datafiles = [x for x in datafiles if
                   x == self._fn_series.format(self.meta_info_id)]
  
      if len(datafiles) != 1:
        raise ValueError("Datafiles content: {}".format(datafiles))
  
      fn = os.path.join(
        self.outer.get_dir(prefix),
        datafiles[0]
      )

      data_file = h5py.File(fn, 'r')
      keys = list(data_file.keys())
      if False:
        self.P("Loaded sparse series keys from '{}':\n{}".format(fn, keys))
      return keys
    #enddef
  
    def _load_sparse_series_subset(self, subset, prefix, sanity=True):
      """
      Generator that yields series data saved as sparse matrices
      and returns a dictionary with each dataset name as a key,
      and the dataset as value
      The file must have the name `prefix`_series_file.hdf5.
      If no prefix is specified, it will use log.file_prefix
  
      Parameters:
      -----------
        1. `prefix`: the filename prefix
        2. `subset`: the subset of the data to be loaded
  
      Returns:
      --------
        `data`: dictionary with datasets name as keys and
        datasets data as values
      """
      
      datafiles = os.listdir(self.outer.get_dir(prefix))
      datafiles = [x for x in datafiles if
                   x == self._fn_series.format(self.meta_info_id)]
  
      if len(datafiles) != 1:
        raise ValueError("Datafiles content: {}".format(datafiles))
      
      fn = os.path.join(
        self.outer.get_dir(prefix),
        datafiles[0]
      )
      if False:
        self.P("Loading sparse series (subset={}) from '{}' ...".format(subset, fn))

      dataset = load_dict_from_hdf5(fn, root=subset)

      if False:
        self.P(" ", t=True)
      
      if sanity:
        self._sanity_check_sparse_dataset(
          dataset,
          target=None,
          nr_series=3
        )
      
      return dataset
    #enddef
    
    def _load_sparse_series(self, prefix, sanity=True):
      keys = self._load_sparse_series_keys(prefix=prefix)
      datasets = []
      
      for k in keys:
        datasets.append(self._load_sparse_series_subset(
          k,
          prefix=prefix,
          sanity=sanity)
        )
      #endfor
      
      return datasets
    #enddef
    
    def _sparse_series_to_array(self, *datasets):
      """
      Converts series from sparse array representation to numpy arrays.
      Each row of the numpy array represents one series while each column
      represents one timestep
  
      Parameters:
      -----------
        1. `datasets`: a dictionary with the stored datasets.
        It must contain the following keys:
          - `row_ind`: the row index for each transaction
          - `col_ind`: the column index for each transaction
          - `mapping` : 
          - `quantity_field`: the quantity sold for each transaction
  
      Returns:
      --------
        1.`np_qty`: quantity series as a numpy array,
        each row is a series, each column is a timestamp
        3.`series_map`: mapping as a list from each row number to series_id
      """
      if False:
        self.P("Converting sparse series to numpy array ...")
      self.log.start_timer('_sparse_series_to_array_loop', section=self.outer._timers_section)
      lst_np_datasets = []
      lst_np_mappings = []
      for dataset in datasets:
        self.log.start_timer('_sparse_series_to_array_dataset', section=self.outer._timers_section)
        mapping = dataset[NumpyizeConstants.SPARSE_MAPPING]
        freq    = dataset[NumpyizeConstants.SPARSE_FREQ]
        
        arrays = {}
        for k in set(dataset.keys()) - set(NumpyizeConstants.SPARSE_DEFAULTS):
          values  = dataset[k][NumpyizeConstants.SPARSE_VALUES]
          row_ind = mapping[dataset[k][NumpyizeConstants.SPARSE_ROW_IND]]
          col_ind = dataset[k][NumpyizeConstants.SPARSE_COL_IND]
          data = (values, (row_ind, col_ind))
          shape = (len(self.outer.series2agg), len(self.dates2id))
          sparse_data = sparse.csr_matrix(data, shape=shape)
          
          np_values = np.squeeze(np.asarray(sparse_data.todense()))
          arrays[k] = np_values
        #endfor
        lst_np_datasets.append(arrays)
        lst_np_mappings.append(mapping)
        self.log.end_timer_no_skip('_sparse_series_to_array_dataset', section=self.outer._timers_section)
      #endfor
      
      self.log.end_timer_no_skip('_sparse_series_to_array_loop', section=self.outer._timers_section)
      if False:
        self.P("", t=True)
      
      assert len(lst_np_datasets) == len(lst_np_mappings)
      return lst_np_datasets, lst_np_mappings, freq
    #enddef
  #endclass Dataset
#endclass DataMaster

if __name__ == '__main__':

  # run it twice; the first time is slower because it has to compute everything;
  # the second time will run faster because everything is cached in the "PREFIX" folder

  from naeural_core import Logger
  # import pandas as pd

  # in our case we will work with the current cache, not dropbox
  log = Logger(lib_name='TST', base_folder='dropbox', app_folder='_local_data/_product_dynamics/_experiments', TF_KERAS=False)

  # set the fields accordingly
  # prefix is the folder in logger's _output where all the sparse series objects and metadata are saved.
  TARGET_FIELD = 'Qtty'
  PREFIX = '20220310'
  eng = TimeseriesDataMaster(
    log=log, prefix=PREFIX,
    aggregation_fields=['ItemId', 'SiteId'],
    time_field='TimeStamp',
    target_field=TARGET_FIELD,
  )

  # in our case, these will be in-memory
  df_train = log.load_dataframe(fn='train.csv', timestamps=['TimeStamp'])
  df_delta = log.load_dataframe(fn='delta.csv', timestamps=['TimeStamp'])

  ds_master = eng.create_dataset()
  ds_delta = eng.create_dataset()

  # ds_master.setup(
  #   column_names=[TARGET_FIELD],
  #   freq='w-mon', # set to 'h'
  #   df=df_train
  # )
  #
  # ds_delta.setup(
  #   column_names=[TARGET_FIELD],
  #   freq='w-mon', # set to 'h'
  #   df=df_delta
  # )

  dct_datasets_master, _, _ = ds_master()
  dct_datasets_delta,  _, _ = ds_delta()

  np_master = dct_datasets_master[TARGET_FIELD]
  np_delta = dct_datasets_delta[TARGET_FIELD]

  np_master.shape # (262933, 157)
  np_delta.shape # (262933, 19) - all the series that are not found at all in the delta dataset are filled with 0!



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Logging utils
"""

import warnings
from threading import Thread

import torch
from torch.utils.tensorboard import SummaryWriter

from utils.general import colorstr, emojis
from utils.loggers.wandb.wandb_utils import WandbLogger
from utils.plots import plot_images, plot_results
from utils.torch_utils import de_parallel

LOGGERS = ('csv', 'tb', 'wandb')  # text-file, TensorBoard, Weights & Biases

try:
    import wandb

    assert hasattr(wandb, '__version__')  # verify package import not local dir
except (ImportError, AssertionError):
    wandb = None


class Loggers():
    # YOLOv5 Loggers class
    def __init__(self, save_dir=None, weights=None, opt=None, hyp=None, logger=None, include=LOGGERS):
        self.save_dir = save_dir
        self.weights = weights
        self.opt = opt
        self.hyp = hyp
        self.logger = logger  # for printing results to console
        self.include = include
        self.keys = ['train/box_loss', 'train/obj_loss', 'train/cls_loss',  # train loss
                     'metrics/precision', 'metrics/recall', 'metrics/mAP_0.5', 'metrics/mAP_0.5:0.95',  # metrics
                     'val/box_loss', 'val/obj_loss', 'val/cls_loss',  # val loss
                     'x/lr0', 'x/lr1', 'x/lr2']  # params
        for k in LOGGERS:
            setattr(self, k, None)  # init empty logger dictionary
        self.csv = True  # always log to csv

        # Message
        if not wandb:
            prefix = colorstr('Weights & Biases: ')
            s = f"{prefix}run 'pip install wandb' to automatically track and visualize YOLOv5 🚀 runs (RECOMMENDED)"
            print(emojis(s))

        # TensorBoard
        s = self.save_dir
        if 'tb' in self.include and not self.opt.evolve:
            prefix = colorstr('TensorBoard: ')
            self.logger.info(f"{prefix}Start with 'tensorboard --logdir {s.parent}', view at http://localhost:6006/")
            self.tb = SummaryWriter(str(s))

        # W&B
        if wandb and 'wandb' in self.include:
            wandb_artifact_resume = isinstance(self.opt.resume, str) and self.opt.resume.startswith('wandb-artifact://')
            run_id = torch.load(self.weights).get('wandb_id') if self.opt.resume and not wandb_artifact_resume else None
            self.opt.hyp = self.hyp  # add hyperparameters
            self.wandb = WandbLogger(self.opt, run_id)
        else:
            self.wandb = None

    def on_pretrain_routine_end(self):
        # Callback runs on pre-train routine end
        paths = self.save_dir.glob('*labels*.jpg')  # training labels
        if self.wandb:
            self.wandb.log({"Labels": [wandb.Image(str(x), caption=x.name) for x in paths]})

    def on_train_batch_end(self, ni, model, imgs, targets, paths, plots, sync_bn):
        # Callback runs on train batch end
        if plots:
            if ni == 0:
                if not sync_bn:  # tb.add_graph() --sync known issue https://github.com/ultralytics/yolov5/issues/3754
                    with warnings.catch_warnings():
                        warnings.simplefilter('ignore')  # suppress jit trace warning
                        self.tb.add_graph(torch.jit.trace(de_parallel(model), imgs[0:1], strict=False), [])
            if ni < 3:
                f = self.save_dir / f'train_batch{ni}.jpg'  # filename
                Thread(target=plot_images, args=(imgs, targets, paths, f), daemon=True).start()
            if self.wandb and ni == 10:
                files = sorted(self.save_dir.glob('train*.jpg'))
                self.wandb.log({'Mosaics': [wandb.Image(str(f), caption=f.name) for f in files if f.exists()]})

    def on_train_epoch_end(self, epoch):
        # Callback runs on train epoch end
        if self.wandb:
            self.wandb.current_epoch = epoch + 1

    def on_val_image_end(self, pred, predn, path, names, im):
        # Callback runs on val image end
        if self.wandb:
            self.wandb.val_one_image(pred, predn, path, names, im)

    def on_val_end(self):
        # Callback runs on val end
        if self.wandb:
            files = sorted(self.save_dir.glob('val*.jpg'))
            self.wandb.log({"Validation": [wandb.Image(str(f), caption=f.name) for f in files]})

    def on_fit_epoch_end(self, vals, epoch, best_fitness, fi):
        # Callback runs at the end of each fit (train+val) epoch
        x = {k: v for k, v in zip(self.keys, vals)}  # dict
        if self.csv:
            file = self.save_dir / 'results.csv'
            n = len(x) + 1  # number of cols
            s = '' if file.exists() else (('%20s,' * n % tuple(['epoch'] + self.keys)).rstrip(',') + '\n')  # add header
            with open(file, 'a') as f:
                f.write(s + ('%20.5g,' * n % tuple([epoch] + vals)).rstrip(',') + '\n')

        if self.tb:
            for k, v in x.items():
                self.tb.add_scalar(k, v, epoch)

        if self.wandb:
            self.wandb.log(x)
            self.wandb.end_epoch(best_result=best_fitness == fi)

    def on_model_save(self, last, epoch, final_epoch, best_fitness, fi):
        # Callback runs on model save event
        if self.wandb:
            if ((epoch + 1) % self.opt.save_period == 0 and not final_epoch) and self.opt.save_period != -1:
                self.wandb.log_model(last.parent, self.opt, epoch, fi, best_model=best_fitness == fi)

    def on_train_end(self, last, best, plots, epoch):
        # Callback runs on training end
        if plots:
            plot_results(file=self.save_dir / 'results.csv')  # save results.png
        files = ['results.png', 'confusion_matrix.png', *[f'{x}_curve.png' for x in ('F1', 'PR', 'P', 'R')]]
        files = [(self.save_dir / f) for f in files if (self.save_dir / f).exists()]  # filter

        if self.tb:
            import cv2
            for f in files:
                self.tb.add_image(f.stem, cv2.imread(str(f))[..., ::-1], epoch, dataformats='HWC')

        if self.wandb:
            self.wandb.log({"Results": [wandb.Image(str(f), caption=f.name) for f in files]})
            # Calling wandb.log. TODO: Refactor this into WandbLogger.log_model
            if not self.opt.evolve:
                wandb.log_artifact(str(best if best.exists() else last), type='model',
                                   name='run_' + self.wandb.wandb_run.id + '_model',
                                   aliases=['latest', 'best', 'stripped'])
                self.wandb.finish_run()
            else:
                self.wandb.finish_run()
                self.wandb = WandbLogger(self.opt)



from concurrent.futures import ThreadPoolExecutor
from threading import Thread
import os


class _ThreadingAPIMixin():
  def __init__(self):
    super(_ThreadingAPIMixin, self).__init__()
    return
  
  def __get_n_threads(self, n_threads):
    if n_threads is None:
      n_threads = 1
    if n_threads > os.cpu_count() // 4 and n_threads > 1:
      self.P("Warning: n_threads is too high, setting to 1/4 of available CPUs", color='y')
      n_threads = max(os.cpu_count() // 4, 1)
    return n_threads
  
  def __wrapper_func_run(self, func: callable, lst_results: list, thread_id: int, n_threads: int):
    lst_results[thread_id] = func(thread_id, n_threads)
    return

  def threadapi_map(self, func, lst_data, n_threads=1):
    """
    Run a function in parallel using ThreadPoolExecutor.map

    Parameters
    ----------
    func : callable
        The function to run in parallel
    lst_data : list
        The list of data to pass to the function
    n_threads : int, optional
        The number of threads to use, by default 1
        If this number is higher than 1/4 of available CPUs, it will be set to 1/4 of available CPUs

    Returns
    -------
    list
        The results of the function (similar to list(map(func, lst_data)))
    """
    n_threads = self.__get_n_threads(n_threads)
    with ThreadPoolExecutor(max_workers=n_threads) as executor:
      results = executor.map(func, lst_data)
    return list(results)

  def threadapi_base64_code_map(self, base64_code, lst_data, n_threads=1):
    """
    Run a custom code method in parallel using ThreadPoolExecutor.map

    Parameters
    ----------
    base64_code : str
        The base64 encoded custom code method
    lst_data : list
        The list of data to pass to the custom code method
    n_threads : int, optional
        The number of threads to use, by default 1
        If this number is higher than 1/4 of available CPUs, it will be set to 1/4 of available CPUs

    Returns
    -------
    list
        The results of the custom code method (similar to list(map(func, lst_data)))
    """
    base_custom_code_method, errors, warnings = self._get_method_from_custom_code(
      str_b64code=base64_code,
      self_var='plugin',
      method_arguments=['plugin', 'data']
    )

    custom_code_method = lambda data: base_custom_code_method(self, data)
    if errors is not None:
      errors_str = "\n".join([str(e) for e in errors])
      self.P(f"Errors found while getting custom code method:\n{errors_str}", color='r')
      return None
    
    warnings_str = "\n".join([str(w) for w in warnings])
    if len(warnings) > 0:
      self.P(f"Warnings found while getting custom code method:\n{warnings_str}", color='y')

    return self.threadapi_map(custom_code_method, lst_data, n_threads)
  
  def threadapi_run(self, func, n_threads):
    """
    Run a function in parallel using threads

    Parameters
    ----------
    func : callable
        The function to run in parallel
        This function must have the following signature: func(thread_id: int, n_threads: int)
    n_threads : int
        The number of threads to use, by default 1
        If this number is higher than 1/4 of available CPUs, it will be set to 1/4 of available CPUs

    Returns
    -------
    list
        A list of results from the function calls, similar to [func(0, n_threads), func(1, n_threads), ... func(n_threads-1, n_threads)]
    """
    # create n_threads threads that run func
    n_threads = self.__get_n_threads(n_threads)

    lst_results = [None] * n_threads

    lst_threads: list[Thread] = []

    for thread_id in range(n_threads):
      thread = Thread(
        target=self.__wrapper_func_run, args=(func, lst_results, thread_id, n_threads),
        daemon=True
      )
      lst_threads.append(thread)

    for thread in lst_threads:
      thread.start()
      
    for thread in lst_threads:
      thread.join()

    return lst_results



#local dependencies
import os
import re
import cv2
import numpy as np

import zipfile

from time import time, sleep
from collections import deque

#global dependencies
from naeural_core import constants as ct

from naeural_core.data.base import DataCaptureThread

_CONFIG = {
  **DataCaptureThread.CONFIG,
  "DELETE_ZIP": True,
  'LOG_PROGRESS_PERIOD': 2000,
  'START_IMAGE_ITER': 0,
  'END_IMAGE_ITER': 0,
  'VALIDATION_RULES' : {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
  },
}

class ImageArchiveDataCapture(DataCaptureThread):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self._path = None
    self._progress_interval = list(range(0, 101, 5))
    self._files = deque()
    self._crt_frame = 0
    self._total_files = 0
    
    super(ImageArchiveDataCapture, self).__init__(**kwargs)
    return
  
  def startup(self):
    super().startup()
    # Following lines have been commented due to limitation of DCT use-cases!
    # assert self.is_keepalive or not self.is_reconnectable
    # assert not self.cfg_live_feed

    self._metadata.update(
      fps=None,
      frame_h=None,
      frame_w=None,
      frame_count=None,
      frame_current=None,
      download=None,
      total_time=None,
      time_current=None,
      frame_msec=None
    )
    return
  
  @property
  def cfg_c_array(self):
    return self.cfg_stream_config_metadata.get('C_ARRAY', True)

  @property
  def cfg_notify_download_downstream(self):
    return self.cfg_stream_config_metadata.get('NOTIFY_DOWNLOAD_DOWNSTREAM', False)

  def _notify_download_progress(self, message):    
    try:
      prc = message.split(':')[1]
      prc = prc.replace('%', '')
      prc = int(float(prc))
      if prc % 5 == 0 and prc in self._progress_interval:
        if prc == 0:
          self._metadata.download_start_date = time()
        
        self._metadata.download = prc
        self._metadata.download_elapsed_time = time() - self._metadata.download_start_date
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_NORMAL,
          msg='Download status',
          info="{}".format(self._metadata.__dict__)
        )
        self._progress_interval.remove(prc)
        
        if self.cfg_notify_download_downstream:
          metadata = self._metadata.__dict__.copy()
          self._add_inputs(
            [
              self._new_input(struct_data=metadata, metadata=metadata),
            ]
          )
    except Exception as e:
      self.P('Exception while notify progress: {}'.format(str(e)), color='r')
    return
  
  def _download_local(self):
    # TODO: unzip file
    self.P(' Found local file.')
    self._path = self.cfg_url
    self._metadata.download = 100
    self._metadata.download_start_date = time()
    self._metadata.download_elapsed_time = time() - self._metadata.download_start_date
    self._create_notification(
      notif=ct.STATUS_TYPE.STATUS_NORMAL,
      msg='Download status',
      info='{}'.format(self._metadata.__dict__)
    )
    if self.cfg_notify_download_downstream:
      metadata = self._metadata.__dict__.copy()
      self._add_inputs(
        [
          self._new_input(struct_data=metadata, metadata=metadata),
        ]
      )
    extract_path = os.path.splitext(self._path)[0]
    
    if self._path.endswith('.zip'):
      with zipfile.ZipFile(self._path, 'r') as fh:
        fh.extractall(extract_path)
      sleep(2)
      if self.cfg_delete_zip:
        os.remove(self._path)
      self._path = extract_path
    self.P(' Path to be used: {}'.format(self._path))
    return

  def _download_cloud(self):
    self.P(' Download...')
    try:
      self._path = self._download(
        url=self.cfg_url,
        progress=False,
        verbose=False,
        notify_download=self._notify_download_progress,
        unzip=True
        )
      self.P(' File downloaded in: {}'.format(self._path))
    except:
      msg = "Exception download"
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        autocomplete_info=True
      )
    return
  
  def _natural_sort(self, l):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    return sorted(l, key=alphanum_key)
  
  def _load_files(self):
    l = []
    for subdir, dirs, files in os.walk(self._path):
      for file in files:
        l.append(os.path.join(subdir, file))
    # endfor os.walk
    self.P('Found {} files in {}'.format(len(l), self._path))
    l = self._natural_sort(l)
    nr_skip_start = self.cfg_start_image_iter
    nr_skip_end = self.cfg_end_image_iter
    if nr_skip_end == 0:
      nr_skip_end = len(l)
    for fn in l[nr_skip_start:nr_skip_end]:
      self._files.append(fn)
    self.P(f'Skipped {nr_skip_start + len(l) - nr_skip_end}({nr_skip_start}+{len(l) - nr_skip_end}) files ({len(self._files)} remaining).')
    self._total_files = len(self._files)
    return
  
  def _init(self):
    self.P("Using '{}' as data source".format(self.cfg_url))
    if os.path.exists(self.cfg_url):
      self.P("  Found '{}' in local".format(self.cfg_url))
      self._download_local()
    else:
      self.P("  '{}' not found in local. Assuming URL".format(self.cfg_url))
      self._download_cloud()
    return
  
  #abstract methods implementation
  def _release(self):
    """Cleaning all capture storage"""
    del self._files
    return
  
  def _maybe_reconnect(self):    
    """Nothing to do for this plugin"""
    if self.has_connection:
      return
    
    self._load_files()
    self._metadata.frame_count = len(self._files)
    self.has_connection = True
    return
  
  def _run_data_aquisition_step(self):
    """Reading images one at a time"""
    if self._files:
      fn = self._files.popleft()
      frame = cv2.imread(fn)
      # frame = np.ascontiguousarray(frame[:,:,::-1])
      frame = frame[:,:,::-1]
      if self.cfg_c_array:
        self.start_timer('ascontiguousarray')
        frame = np.ascontiguousarray(frame)
        self.end_timer('ascontiguousarray')
      #endif

      #metadata
      self._crt_frame+= 1

      if self._crt_frame % self.cfg_log_progress_period == 0:
        self.P(f'So far added {self._crt_frame}/{self._total_files} to inputs queue.')

      height, width, _ = frame.shape
      self._metadata.frame_h = height
      self._metadata.frame_w = width
      self._metadata.frame_current = self._crt_frame
      self._metadata.subdir = os.path.relpath(os.path.dirname(fn), start=self._path)

      self._add_inputs(
        [
          self._new_input(img=frame, metadata=self._metadata.__dict__.copy()),
        ]
      )
    else:
      self._crt_frame = 0
      self.has_connection = False
      self.has_finished_acquisition = True
    return


"""
[THTST][2021-10-06 14:51:47] Timing results:
[THTST][2021-10-06 14:51:47]  cuda_fp32 = 0.0003s, max: 0.0010s, curr: 0.0010s
[THTST][2021-10-06 15:10:06]  cuda_fp32 = 0.0016s, max: 0.0070s, curr: 0.0009s
[THTST][2021-10-06 15:14:06]  cuda_fp32 = 0.0015s, max: 0.0070s, curr: 0.0010s

[THTST][2021-10-06 15:14:07] 0.001545557627329478

Script:
[THTST][2021-10-06 15:15:07]  torch_script = 0.0015s, max: 0.0060s, curr: 0.0010s
[THTST][2021-10-06 15:15:39]  torch_script = 0.0015s, max: 0.0070s, curr: 0.0021s
[THTST][2021-10-06 15:15:39] 0.001549411225724626
[THTST][2021-10-06 15:15:07] 0.0015447041890523335



[THTST][2021-10-06 14:51:47]  cuda_fp16 = 0.0011s, max: 0.0050s, curr: 0.0030s

"""
import numpy as np
import torch as th
import torchvision as tv

from naeural_core import Logger
from naeural_core.local_libraries.nn.th.trainer import ModelTrainer
from naeural_core.local_libraries.nn.th.layers import InputPlaceholder
from naeural_core.local_libraries.nn.th.utils import conv_output_shape

  

if __name__ == '__main__':
  
  class ConvBlock(th.nn.Module):
    def __init__(self, input_shape, k, f, s , act):
      super().__init__()
      prev_ch, prev_h, prev_w = input_shape
      self.conv = th.nn.Conv2d(
        in_channels=prev_ch,
        out_channels=f,
        kernel_size=k,
        stride=s)
      self.bn = th.nn.BatchNorm2d(num_features=f)
      self.output_shape = conv_output_shape(
          h_w=(prev_h, prev_w),
          kernel_size=k,
          stride=s,
          )
      self.act = th.nn.LeakyReLU()
      return
    
    def forward(self, inputs):
      th_x = self.conv(inputs)
      th_x = self.bn(th_x)
      th_x = self.act(th_x)
      return th_x

  class SimpleModel(th.nn.Module):
    def __init__(self, 
                 input_shape=(1, 28, 28), 
                 convs=[(8, 3, 2), (16, 3, 2), (32, 3, 1)], 
                 dense=32,
                 **kwargs):
      super().__init__()
      assert len(input_shape) == 3
      self.layers = th.nn.ModuleList()
      self.layers.append(InputPlaceholder(input_shape))
      input_size = input_shape[0]
      prev_h, prev_w = input_shape[1:]
      for i, (f, k, s) in enumerate(convs):
        cnv = th.nn.Conv2d(
          in_channels=input_size, 
          out_channels=f, 
          kernel_size=k,
          stride=s
          )
        input_size = f
        new_h, new_w = conv_output_shape(
          h_w=(prev_h, prev_w),
          kernel_size=k,
          stride=s,
          )
        self.layers.append(cnv)
        self.layers.append(th.nn.ReLU6())
        prev_h = new_h
        prev_w = new_w

      self.layers.append(th.nn.Flatten())
      post_flatten_size = new_h * new_w * f
      self.layers.append(th.nn.Linear(post_flatten_size, dense))
      self.layers.append(th.nn.ReLU6())
      self.layers.append(th.nn.Dropout(0.5))
      self.layers.append(th.nn.Linear(dense, 10))
      return
    
    def forward(self, inputs):
      x = inputs
      for layer in self.layers:
        x = layer(x)
      return x

  

  l = Logger('THTST', base_folder='.', app_folder='_cache', TF_KERAS=False)
  
  train = tv.datasets.MNIST(root=l.get_data_folder(), train=True, download=True)
  test = tv.datasets.MNIST(root=l.get_data_folder(), train=False, download=True)
  
  
  (x_train, y_train), (x_dev, y_dev) = (train.data.numpy(), train.targets.numpy()), (test.data.numpy(), test.targets.numpy())
  x_train = (x_train.reshape((-1, 1, 28, 28)) / 255.).astype('float32')
  x_dev = (x_dev.reshape((-1, 1, 28, 28)) / 255.).astype('float32')
  y_train = y_train.astype('int64')
  y_dev = y_dev.astype('int64')
  
  l.gpu_info(True)
  
  model = SimpleModel()
  
  if False:
    model_loss = th.nn.CrossEntropyLoss()
    
    trainer = ModelTrainer(
      log=l, 
      model=model, 
      losses=model_loss, 
      validation_data=(x_dev, y_dev),
      batch_size=256
      )
    trainer.fit(x_train, y_train)
  else:  
    dev = th.device('cuda')
    model.to(dev)
    model.eval()
    l.P("Model: \n{}".format(model))
    th_x_dev = th.tensor(x_dev, device=dev)
    with th.no_grad():
      timings = []
      for _ in range(1000):
        l.start_timer('cuda_fp32')
        y_pred = model(th_x_dev)
        tt = l.stop_timer('cuda_fp32')
        timings.append(tt)
        
      if False:
        th_x_dev = th_x_dev.half()
        model.half()
        for _ in range(1000):
          l.start_timer('cuda_fp16')
          y_pred = model(th_x_dev)
          l.stop_timer('cuda_fp16')
          
    l.show_timers()
    np_yh = y_pred.cpu().numpy()
    l.save_pickle_to_data((x_dev, np_yh), 'test_mnist0')
    scr_model = th.jit.script(model)
    th.jit.save(scr_model, 'xperimental/th_basic/mnist0.sth')
    
    np_t = np.array(timings[1:])
    l.P(np_t.mean())
  
  
  



class _WorkingHoursMixin(object):

  def __init__(self):
    self.__is_new_shift = False
    self.__shift_started = False
    super(_WorkingHoursMixin, self).__init__()
    return

  def _valid_hour_interval(self, interval):
    """
    Method for validating an interval.
    Parameters
    ----------
    interval : list - list of 2 strings in format HH:MM

    Returns
    -------
    res - bool - True if the interval is valid, False otherwise
    """
    if not isinstance(interval, list) or len(interval) != 2:
      return False
    if not all(isinstance(time, str) for time in interval):
      return False
    for time in interval:
      if len(time.split(':')) < 2:
        return False
      hour, minute = time.split(':')[:2]
      if not (0 <= int(hour) < 24 and 0 <= int(minute) < 60):
        return False
    return True

  def _valid_hour_schedule(self, hour_schedule):
    """
    Method for validating the hour schedule.
    Parameters
    ----------
    hour_schedule : list - list of intervals described as a list of 2 strings in format HH:MM

    Returns
    -------
    res - bool - True if the schedule is valid, False otherwise
    """
    if not isinstance(hour_schedule, list):
      return False
    if not all(self._valid_hour_interval(interval) for interval in hour_schedule):
      return False
    return True

  def validate_working_hours(self):
    """
    Method for validating the working hours configuration.
    Returns
    -------
    res - bool - True if the configuration is valid, False otherwise
    """
    working_hours = self.cfg_working_hours
    if working_hours is None:
      return True
    err_message = (
      f"Invalid working hours configuration for plugin {self}.\nThe configuration "
      f"should be a list of intervals, with each interval described as a list of 2 "
      f"strings in format \"HH:MM\" or a dict with the days as keys and the "
      f"already described list format as values.\n"
    )
    is_valid = True
    if isinstance(working_hours, list):
      # if the schedule is a list of intervals we consider it to be a universal schedule(the weekday is irrelevant)
      if not self._valid_hour_schedule(working_hours):
        is_valid = False
    elif isinstance(working_hours, dict):
      # if the schedule is a dict we consider it to be a schedule based on week days and besides the check made for the
      # intervals we also check if the keys are valid week days.
      if not all(
        key.upper() in self.ct.WEEKDAYS_SHORT and isinstance(value, list) and self._valid_hour_schedule(value)
        for key, value in working_hours.items()
      ):
        is_valid = False
    else:
      # if the schedule is not a list or a dict it is invalid.
      is_valid = False
    # endif working_hours type
    if not is_valid:
      self.add_error(err_message)
    return is_valid

  def get_timezone(self):
    tz = self.cfg_working_hours_timezone
    if tz is None:
      return self.log.timezone
    return tz

  def str_to_datetime(self, str_time, weekday=None):
    """
    Convert a string time to a datetime object
    Parameters
    ----------
    str_time : str - time in format HH:MM
    weekday : int or None - the weekday index starting from 0

    Returns
    -------
    datetime object with the time set to the provided one
    """
    # This will be 1st of January 1900, Monday
    res_dt = self.datetime.strptime(str_time, '%H:%M')
    if weekday is not None:
      res_dt = res_dt.replace(day=weekday + 1)
    # This is done in order to avoid invalid argument errors when converting the datetime to local time
    res_dt = res_dt + self.timedelta(weeks=5000)
    return res_dt

  def interval_to_local(self, interval, weekday=None, timezone=None):
    """
    Method for converting an interval to local time.
    In case the weekday is provided and the interval is crossing the midnight
    this will return a list of 2 intervals, one for each day.
    Parameters
    ----------
    interval : list - list of 2 strings representing the start and end of the interval in format HH:MM
    weekday : int or None - the weekday index starting from 0
    timezone : str or None - the timezone to convert to

    Returns
    -------
    res - list of 1 or 2 tuples representing the weekday, start and end of the interval in local time.
    """
    res = []
    start_dt = self.str_to_datetime(interval[0], weekday)
    end_dt = self.str_to_datetime(interval[1], weekday)
    tz = timezone if timezone is not None else self.get_timezone()
    start_dt_local = self.log.utc_to_local(start_dt, tz)
    end_dt_local = self.log.utc_to_local(end_dt, tz)
    if start_dt_local.weekday() != end_dt_local.weekday():
      res.append((start_dt_local.weekday(), start_dt_local, self.datetime.combine(start_dt_local.date(), self.datetime.max.time())))
      res.append((end_dt_local.weekday(), self.datetime.combine(end_dt_local.date(), self.datetime.min.time()), end_dt_local))
    else:
      res.append((start_dt_local.weekday(), start_dt_local, end_dt_local))
    return res

  def working_hours_to_local(self, working_hours_schedule, timezone=None):
    """
    Method for converting the working hours to local time.
    Parameters
    ----------
    working_hours_schedule : list or dict - the working hours schedule
    timezone : str or None - the timezone to convert to

    Returns
    -------
    res_working_hours - list or dict with the working hours (and weekdays if necessary) converted to local time
    """
    dct_working_hours = {}
    universal_schedule = False
    if isinstance(working_hours_schedule, list):
      # if the schedule is a list of intervals we consider it to be a universal schedule(the weekday is irrelevant)
      universal_schedule = True
      dct_working_hours = {None: working_hours_schedule}
    elif isinstance(working_hours_schedule, dict):
      dct_working_hours = {**working_hours_schedule}
    # endif list
    tz = timezone if timezone is not None else self.get_timezone()
    lst_all_intervals = []
    for key, lst_intervals in dct_working_hours.items():
      if key is None or key.upper() in self.ct.WEEKDAYS_SHORT:
        weekday_idx = None if key is None else self.ct.WEEKDAYS_SHORT.index(key.upper())
        for interval in lst_intervals:
          lst_all_intervals.extend(self.interval_to_local(interval, weekday_idx, timezone=tz))
        # endfor
      # endif key is None or key in WEEKDAYS_SHORT
    # endfor
    if universal_schedule:
      res_working_hours = [
        [start_dt.strftime('%H:%M'), end_dt.strftime('%H:%M')]
        for _, start_dt, end_dt in lst_all_intervals
      ]
    else:
      res_working_hours = {}
      for (key, start_dt, end_dt) in lst_all_intervals:
        key_str = self.ct.WEEKDAYS_SHORT[key]
        if key_str not in res_working_hours:
          res_working_hours[key_str] = []
        res_working_hours[key_str].append([start_dt.strftime('%H:%M'), end_dt.strftime('%H:%M')])
      # endfor
    # endif universal_schedule
    return res_working_hours

  @property
  def working_hours(self):
    schedule = self.cfg_working_hours

    # adapted also in case of having just one interval and mistakenly the operator did not configured list of lists.
    if isinstance(schedule, list) and len(schedule) == 2 and isinstance(schedule[0], str):
      schedule = [schedule]

    if isinstance(schedule, dict):
      schedule = {
        key.upper(): value
        for key, value in schedule.items()
      }
    # endif dict
    try:
      res_working_hours = self.working_hours_to_local(schedule, timezone=self.get_timezone())
    except Exception as e:
      self.P("Exception {} occurent in working_hours_to_local:\ncfg_working_hours:\n{}\n\nschedule:\n{}".format(
        e, self.json_dumps(self.cfg_working_hours), self.json_dumps(schedule),
      ))
      raise
    #endtry
    return res_working_hours
  
  
  def on_shift_start(self, interval_idx=None, weekday_name=None, **kwargs):
    """
    This method should be defined in plugins with specific code that will be
    run when a new working hours shift starts.
    """
    return

  def on_shift_end(self, **kwargs):
    """
    This method should be defined in plugins with specific code that will be
    run when a working hours shift ends.
    """
    return

  def __on_shift_start(self, interval_idx, weekday_name=None):
    # TODO: in future implement dict params for instance config in a specific time interval
    hour_schedule = self.working_hours[weekday_name] if weekday_name is not None else self.working_hours
    hrs = 'NON-STOP' if interval_idx is None else hour_schedule[interval_idx]
    shift_str = f'{hrs}' if weekday_name is None else f'{weekday_name}: {hrs}[{self.get_timezone()}]'
    msg = f"Starting new working hours shift {shift_str}"
    info = f"Plugin {self} starting new shift {shift_str}"
    self.P(msg)
    self._create_notification(
      msg=msg,
      info=info,
      displayed=True,
      notif_code=self.ct.NOTIFICATION_CODES.PLUGIN_WORKING_HOURS_SHIFT_START, 
    )
    self.add_payload_by_fields(
      status="Working hours shift STARTED",
      working_hours=self.cfg_working_hours,
      working_hours_timezone=self.get_timezone(),
      forced_pause=self.cfg_forced_pause,
      ignore_working_hours=self.cfg_ignore_working_hours,
      img=None,
    )
    self.P("Executing on_shift_start method")
    self.on_shift_start(interval_idx=interval_idx, weekday_name=weekday_name)
    return

  def __on_shift_end(self):
    msg = f"Ended current working hours shift for {self}. The full schedule is: {self.working_hours}[{self.get_timezone()}]"
    info = f"Plugin {self} ended its shift shift. The full schedule is: {self.working_hours}[{self.get_timezone()}]"
    self.P(msg, color='r')
    self._create_notification(
      msg=msg,
      info=info,
      working_hours=self.cfg_working_hours,
      working_hours_timezone=self.get_timezone(),
      forced_pause=self.cfg_forced_pause,
      displayed=True,
      notif_code=self.ct.NOTIFICATION_CODES.PLUGIN_WORKING_HOURS_SHIFT_END, 
    )
    self.add_payload_by_fields(
      status="Working hours shift ENDED",
      working_hours=self.cfg_working_hours,
      working_hours_timezone=self.get_timezone(),
      ignore_working_hours=self.cfg_ignore_working_hours,
      forced_pause=self.cfg_forced_pause,
      img=None,
    )
    self.P("Executing on_shift_end method")
    self.on_shift_end()
    return

  def __get_outside_working_hours(self):
    interval_idx = None
    result = True
    weekday_name = None

 # if the plugin is configured to ignore working hours it will always be considered as inside working hours
    if self.cfg_ignore_working_hours:      
      return False, interval_idx, weekday_name
    
    # if the provided working_hours is None the plugin will always be outside working hours
    if self.working_hours is None:
      return True, interval_idx, weekday_name

    ts_now = self.datetime.now()
    # extracting both the weekday and the hour intervals if it's the case
    lst_hour_schedule, weekday_name = self.log.extract_weekday_schedule(
      ts=ts_now,
      schedule=self.working_hours,
      return_day_name=True
    )
    
    # in case we have schedule based on week days and the current day was not specified
    # it means we are outside the working hours
    
    if lst_hour_schedule is not None:
      # if hour_schedule is an empty list we have 2 cases:
      # 1. The plugin will work non-stop on the current day (if schedule is using week days)
      # 2. The plugin will work non-stop regardless of the week day
      if len(lst_hour_schedule) == 0:
        result = False

      interval_idx = self.log.extract_hour_interval_idx(
        ts=ts_now,
        lst_schedule=lst_hour_schedule
      )
    # endif hour_schedule is not None
    return result, interval_idx, weekday_name
  

  @property
  def outside_working_hours(self):
    result, interval_idx, weekday_name = self.__get_outside_working_hours()

    # interval found or non-stop functioning
    if interval_idx is not None or not result:
      self.__is_new_shift = False
      if not self.__shift_started:
        self.__is_new_shift = True      
        self.__on_shift_start(
          weekday_name=weekday_name,
          interval_idx=interval_idx
        )
      # endif mark new shift
      self.__shift_started = True
      result = False
    elif self.__shift_started:
      # shift already started so we close it
      self.__on_shift_end()
      self.__shift_started = False
    # endif current time in valid interval
    return result


  @property
  def working_hours_is_new_shift(self):
    return self.__is_new_shift




if __name__ == '__main__':
  from naeural_core import Logger
  from datetime import datetime

  log = Logger(
    'gigi',
    base_folder='.',
    app_folder='_local_cache'
  )
  class Base(object):
    def __init__(self, **kwargs):
      return

  def str_to_datetime(str_time, weekday=None):
    """
    Convert a string time to a datetime object
    Parameters
    ----------
    str_time : str - time in format HH:MM
    weekday : int or None - the weekday index starting from 0

    Returns
    -------
    datetime object with the time set to the provided one
    """
    # This will be 1st of January 1900, Monday
    res_dt = datetime.strptime(str_time, '%H:%M')
    if weekday is not None:
      res_dt = res_dt.replace(day=weekday + 1)
    return res_dt


  def interval_to_datetime(interval):
    return [
      log.str_to_datetime(interval[0]),
      log.str_to_datetime(interval[1])
    ]

  class P(_WorkingHoursMixin, Base):
    def __init__(self, **kwargs):
      super(P, self).__init__(**kwargs)
      return


  working_hours_tests = [
    [
      ['21:10', '10:10'],
      ['11:10', '11:25']
    ],
    [['10:10', '21:10']]
  ]

  for hours in working_hours_tests:
    log.P(hours)
    p = P()
    p.cfg_working_hours = hours
    p.log = log
    log.P(f'`p.outside_working_hours`={p.outside_working_hours}')
    log.P(f'`p.working_hours_is_new_shift`={p.working_hours_is_new_shift}')
    # print(p.outside_working_hours)
    # print(p.working_hours_is_new_shift)




from naeural_core import Logger
from time import sleep

if __name__ == '__main__':
  log1 = Logger(
    lib_name='TST1',
    base_folder='.',
    app_folder='_cache',
    max_lines=1000, 
    TF_KERAS=False
  )

  for i in range(5):
    log1.start_timer('test1',section='test')
    sleep(0.01)
    log1.stop_timer('test1',section='test')
    log1.start_timer('test2',section='test')
    sleep(0.01)
    log1.stop_timer('test2',section='test')
    
  dct_t, dct_tg = log1.export_timers_section(section='test')
  log1.show_timers()  
    
  log = Logger(
    lib_name='TST',
    base_folder='.',
    app_folder='_cache',
    max_lines=1000, 
    TF_KERAS=False
  )
  
  for i in range(5):
    log.start_timer('test1',section='test')
    log.start_timer('test11',section='test')    
    sleep(0.01)
    log.stop_timer('test11',section='test')
    log.start_timer('test12',section='test')    
    sleep(0.01)
    log.stop_timer('test12',section='test')
    log.stop_timer('test1',section='test')
    log.start_timer('test2',section='test')
    sleep(0.01)
    log.stop_timer('test2',section='test')
  
  log.import_timers_section(
    dct_timers=dct_t, 
    dct_timers_graph=dct_tg, 
    section='test_imported'
  )
  
  for i in range(10):
    log.start_timer('main')
    if True:
      log.start_timer('op1')
      if i % 2 == 0:
        log.start_timer('op2')
        #concrete op
        sleep(0.001)
        log.stop_timer('op2')
      if i == 1 or i == 3:
        log.start_timer('op3')
        #concrete op
        sleep(0.001)
        # if i == 1:
        #   continue
        log.stop_timer('op3')        
      if i == 2:
        log.start_timer('op4')        
        #concrete op
        sleep(0.001)
        log.stop_timer('op4')
      log.stop_timer('op1')      
    log.stop_timer('main')
  
  log.show_timers()
  log.P("Overwriting ...", color='m')
  log.import_timers_section(
    dct_timers=dct_t, 
    dct_timers_graph=dct_tg, 
    section='test',
    overwrite=True,
  )
  log.show_timers()
  